"""Interactive module tests."""
from datetime import datetime, timedelta
from unittest import mock

import pytest
from ClusterShell.MsgTree import MsgTreeElem
from cumin import Config, NodeSet
from cumin.transports import Target, clustershell

from spicerack import confctl, remote
from spicerack.tests import get_fixture_path


def mock_cumin(mocked_transports, retcode, retvals=None):
    """Given a mocked cumin.transports, add the necessary mocks for these tests and set the retcode."""
    if retvals is None:
        retvals = [[("host1", b"output1")]]

    results = []
    for retval in retvals:
        result = []
        for host, message in retval:
            result.append((NodeSet(host), MsgTreeElem(message, parent=MsgTreeElem())))

        results.append(result)

    mocked_transports.clustershell = clustershell
    mocked_execute = mock.Mock()
    mocked_execute.return_value = retcode
    mocked_get_results = mock.Mock()
    if results:
        mocked_get_results.side_effect = results
    else:
        mocked_get_results.return_value = iter(())

    mocked_transports.clustershell.ClusterShellWorker.execute = mocked_execute
    mocked_transports.clustershell.ClusterShellWorker.get_results = mocked_get_results
    mocked_transports.Target = Target


class TestRemoteHostsAdapter:
    """Test class for the RemoteHostsAdapter class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        config = get_fixture_path("remote", "config.yaml")
        self.hosts = NodeSet("host[1-9]")
        self.remote_hosts = remote.RemoteHostsAdapter(remote.RemoteHosts(config, self.hosts, dry_run=False))

    def test_str(self):
        """The str() of an instance should return the string representation of the target hosts."""
        assert str(self.remote_hosts) == str(self.hosts)

    def test_len(self):
        """The len() of an instance should return the number of target hosts."""
        assert len(self.remote_hosts) == len(self.hosts)


class TestLBRemoteCluster:
    """Test class for the LBRemoteCluster class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        config = get_fixture_path("remote", "config.yaml")
        self.hosts = NodeSet("host[1-10]")
        # We want to mock out ConftoolEntity completely here. As far as we're concerned it's just an interface
        self.conftool = mock.MagicMock(spec=confctl.ConftoolEntity)
        self.remote_hosts = remote.RemoteHosts(config, self.hosts, dry_run=False)
        self.remote_hosts.run_async = mock.MagicMock()
        self.lbcluster = remote.LBRemoteCluster(config, self.remote_hosts, self.conftool)

    @pytest.mark.parametrize("size", [0, 10])
    def test_run_wrong_batch_size(self, size):
        """Test run fails with a bad batch size."""
        with pytest.raises(remote.RemoteError, match="Values for batch_size"):
            self.lbcluster.run("some command", batch_size=size)

    def test_run_no_depool(self):
        """Test a run with no service to depool."""
        self.lbcluster.run("some command", "some_other cmd")
        self.remote_hosts.run_async.assert_called_with(
            "some command",
            "some_other cmd",
            success_threshold=1.0,
            batch_size=1,
            batch_sleep=None,
            is_safe=False,
            print_output=True,
            print_progress_bars=True,
        )

    def test_run_no_depool_failures(self):
        """Test a run with no service to depool where we allow failures."""
        self.lbcluster.run("some command", "some_other cmd", max_failed_batches=1)
        self.remote_hosts.run_async.assert_called_with(
            "some command",
            "some_other cmd",
            success_threshold=0.9,
            batch_size=1,
            batch_sleep=None,
            is_safe=False,
            print_output=True,
            print_progress_bars=True,
        )

    @mock.patch("spicerack.remote.RemoteHosts.run_async")
    def test_run_depool(self, run_async):
        """Test a run with services to depool."""
        self.conftool.change_and_revert = mock.MagicMock()
        run_async.return_value = [(NodeSet("host1"), None), (NodeSet("host2"), None)]
        res = self.lbcluster.run(
            "test -d /tmp",
            svc_to_depool=["service1", "service2"],
            batch_size=5,
        )

        # Run has been sliced in two
        assert run_async.call_count == 2
        assert res == [
            (NodeSet("host1"), None),
            (NodeSet("host2"), None),
            (NodeSet("host1"), None),
            (NodeSet("host2"), None),
        ]
        # The depool is called on subsequent groups of servers.
        self.conftool.change_and_revert.assert_called_with(
            "pooled",
            "yes",
            "no",
            service="service1|service2",
            name="host6|host7|host8|host9|host10",
        )

    @mock.patch("spicerack.remote.RemoteHosts.run_async")
    def test_run_depool_failure(self, run_async):
        """Test a run with services to depool where a failure is caused."""
        # Case 1: run_async fails, no max_failed_batches
        run_async.side_effect = [
            [(NodeSet("host1"), None)],
            remote.RemoteExecutionError(message="foobar!", retcode=10),
            remote.RemoteExecutionError(message="barbaz!", retcode=10),
        ]
        with pytest.raises(remote.RemoteClusterExecutionError, match="1 hosts have failed execution") as err:
            self.lbcluster.run(
                "test -d /tmp",
                svc_to_depool=["service1", "service2"],
                batch_size=3,
            )
            assert len(err.results) == 1
            assert len(err.failures) == 1
        # This time, we bailed out after the first failure
        assert run_async.call_count == 2

        run_async.side_effect = [
            [(NodeSet("host1"), None)],
            remote.RemoteExecutionError(message="foobar!", retcode=10),
            remote.RemoteExecutionError(message="barbaz!", retcode=10),
        ]
        run_async.reset_mock()
        with pytest.raises(remote.RemoteClusterExecutionError, match="2 hosts have failed execution") as err:
            self.lbcluster.run(
                "test -d /tmp",
                svc_to_depool=["service1", "service2"],
                batch_size=3,
                max_failed_batches=1,
            )
            assert len(err.results) == 1
            assert len(err.failures) == 2
        # All batches have been run, as we could tolerate one error
        assert run_async.call_count == 3

    @mock.patch("spicerack.remote.RemoteHosts.run_async")
    def test_run_depool_no_sleep(self, _):
        """Test a run with services to depool where a failure is caused."""
        with mock.patch("time.sleep") as ts:
            self.lbcluster.run(
                "test -d /tmp",
                svc_to_depool=["service1", "service2"],
                batch_size=3,
            )
            assert ts.call_count == 0

    @mock.patch("spicerack.remote.RemoteHosts.run_async")
    def test_run_depool_sleep(self, _):
        """Test a run where we have a batch sleep."""
        with mock.patch("time.sleep") as ts:
            self.lbcluster.run(
                "test -d /tmp",
                svc_to_depool=["service1", "service2"],
                batch_sleep=3,
                batch_size=3,
            )
            assert ts.call_count == 4

    def test_reload_services(self):
        """Test a service reload."""
        self.lbcluster.run = mock.MagicMock(return_value="foobar")
        assert self.lbcluster.reload_services(["svc"], ["lbl1", "lbl2"]) == "foobar"
        self.lbcluster.run.assert_called_with(
            'systemctl reload "svc"',
            svc_to_depool=["lbl1", "lbl2"],
            batch_size=1,
            batch_sleep=None,
            is_safe=False,
            print_output=True,
            print_progress_bars=True,
        )

    def test_restart_services(self):
        """Test a service restart."""
        self.lbcluster.run = mock.MagicMock(return_value="foobar")
        assert self.lbcluster.restart_services(["svc1", "svc2"], ["lbl1", "lbl2"]) == "foobar"
        self.lbcluster.run.assert_called_with(
            'systemctl restart "svc1"',
            'systemctl restart "svc2"',
            svc_to_depool=["lbl1", "lbl2"],
            batch_size=1,
            batch_sleep=None,
            is_safe=False,
            print_output=True,
            print_progress_bars=True,
        )


class TestRemote:
    """Test class for the Remote class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        config = get_fixture_path("remote", "config.yaml")
        self.remote = remote.Remote(config)

    def test_query_ok(self):
        """Calling query() should return the matching hosts."""
        query = "host[1-9]"
        remote_hosts = self.remote.query(query)

        assert isinstance(remote_hosts, remote.RemoteHosts)
        assert remote_hosts.hosts == NodeSet(query)
        assert str(remote_hosts) == "host[1-9]"
        assert len(remote_hosts) == 9

    def test_query_accepts_sudo(self):
        """Calling query() should return the matching hosts even if using sudo."""
        query = "host[1-9]"
        remote_hosts = self.remote.query(query, use_sudo=True)

        assert isinstance(remote_hosts, remote.RemoteHosts)
        assert remote_hosts.hosts == NodeSet(query)
        assert str(remote_hosts) == "host[1-9]"
        assert len(remote_hosts) == 9
        assert remote_hosts._use_sudo  # pylint: disable=protected-access

    def test_query_invalid(self):
        """Calling query() with an invalid query should raise RemoteError."""
        with pytest.raises(remote.RemoteError, match="Failed to execute Cumin query"):
            self.remote.query("or invalid")

    def test_query_confctl_ok(self):
        """Succesful query_confctl() should return the correct lbremotehosts instance."""
        conftool = mock.MagicMock(spec=confctl.ConftoolEntity)
        host1 = mock.MagicMock()
        host1.name = "host1"
        host2 = mock.MagicMock()
        host2.name = "host2"
        conftool.get.return_value = [host1, host2]
        lbcluster = self.remote.query_confctl(conftool, dc="a", sometag="someval")
        conftool.get.assert_called_with(dc="a", sometag="someval")
        assert str(lbcluster) == "host[1-2]"
        assert isinstance(lbcluster, remote.LBRemoteCluster)

    def test_query_confctl_error(self):
        """Failing query_confctl() raises a RemoteError."""
        conftool = mock.MagicMock(spec=confctl.ConftoolEntity)
        conftool.get.side_effect = confctl.ConfctlError("test test!")
        with pytest.raises(remote.RemoteError, match="Failed to execute the conftool query"):
            self.remote.query_confctl(conftool, dc="a", sometag="someval")


class TestRemoteHosts:
    """Test class for the RemoteHosts class."""

    @mock.patch("spicerack.remote.transports", autospec=True)
    def setup_method(self, _, mocked_transports):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path("remote", "config.yaml"))
        self.mocked_transports = mocked_transports
        self.hosts = NodeSet("host[1-9]")
        self.remote_hosts = remote.RemoteHosts(self.config, self.hosts, dry_run=False)
        self.remote_hosts_dry_run = remote.RemoteHosts(self.config, self.hosts)
        self.expected = [(NodeSet("host1"), "output1")]

    def test_init_no_hosts(self):
        """Should raise RemoteError if initialized without hosts."""
        with pytest.raises(remote.RemoteError, match="No hosts provided"):
            remote.RemoteHosts(self.config, NodeSet(), dry_run=False)

    @pytest.mark.parametrize("func_name", ("run_sync", "run_async"))
    def test_execute(self, func_name):
        """Calling execute() should run the given commands in the target hosts."""
        mock_cumin(self.mocked_transports, 0)
        results = getattr(self.remote_hosts, func_name)("command1")
        assert [(host, msg.message().decode()) for host, msg in results] == self.expected

    @pytest.mark.parametrize("func_name", ("run_sync", "run_async"))
    def test_execute_fail(self, func_name):
        """Calling execute() should raise RemoteError if the Cumin execution fails."""
        mock_cumin(self.mocked_transports, 11)
        with pytest.raises(
            remote.RemoteExecutionError,
            match=r"Cumin execution failed \(exit_code=11\)",
        ) as exc_info:
            getattr(self.remote_hosts, func_name)("command1")

        assert exc_info.value.retcode == 11
        self.mocked_transports.clustershell.ClusterShellWorker.execute.assert_called_once_with()

    @pytest.mark.parametrize("func_name", ("run_sync", "run_async"))
    def test_execute_dry_run_safe(self, func_name):
        """Calling execute() in dry_run mode should run the given commands if marked safe."""
        mock_cumin(self.mocked_transports, 11)  # Simulate a failure
        # In DRY-RUN when is_safe=True all executions are considered successful.
        results = getattr(self.remote_hosts_dry_run, func_name)("command1", is_safe=True)
        assert [(host, msg.message().decode()) for host, msg in results] == self.expected

    @pytest.mark.parametrize("func_name", ("run_sync", "run_async"))
    def test_execute_dry_run_unsafe(self, func_name):
        """Calling execute() in dry_run mode should not run the given commands, considered unsafe by default."""
        results = getattr(self.remote_hosts_dry_run, func_name)("command1")
        assert list(results) == []  # pylint: disable=use-implicit-booleaness-not-comparison

    @pytest.mark.parametrize("func_name", ("run_sync", "run_async"))
    def test_execute_batch_size(self, func_name):
        """Calling execute() with batch_size should parse it to detect percentage or absolute value."""
        mock_cumin(self.mocked_transports, 0)
        getattr(self.remote_hosts, func_name)("command1", batch_size=2)
        # TODO: remove this test once the logic has been moved to Cumin itself

    @mock.patch("spicerack.remote.transports.Target")
    def test_reboot_single(self, mocked_target):
        """It should call the reboot script on the target host with default batch size and no sleep."""
        hosts = NodeSet("host1")
        remote_hosts = remote.RemoteHosts(self.config, hosts, dry_run=False)
        mock_cumin(self.mocked_transports, 0)
        remote_hosts.reboot()
        self.mocked_transports.clustershell.ClusterShellWorker.execute.assert_called_once_with()
        mocked_target.assert_has_calls([mock.call(hosts, batch_size_ratio=None, batch_sleep=None, batch_size=1)])

    @mock.patch("spicerack.remote.transports.Target")
    def test_reboot_many(self, mocked_target):
        """It should call the reboot script on the target hosts with the given batch size and sleep."""
        mock_cumin(self.mocked_transports, 0)
        self.remote_hosts.reboot(batch_size=2, batch_sleep=30.0)
        self.mocked_transports.clustershell.ClusterShellWorker.execute.assert_called_once_with()
        mocked_target.assert_has_calls([mock.call(self.hosts, batch_size_ratio=None, batch_sleep=30.0, batch_size=2)])

    @mock.patch("spicerack.remote.RemoteHosts.uptime")
    def test_wait_reboot_since_ok(self, mocked_uptime):
        """It should return immediately if the host has already a small enough uptime."""
        since = datetime.utcnow() - timedelta(minutes=5)
        mocked_uptime.return_value = [(self.hosts, 30.0)]
        self.remote_hosts.wait_reboot_since(since)
        mocked_uptime.assert_called_once_with(print_progress_bars=True)

    @pytest.mark.parametrize(
        "side_effect",
        (
            remote.RemoteExecutionError(message="unable to connect", retcode=1),
            remote.RemoteError("Unable to extract data"),
        ),
    )
    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    @mock.patch("spicerack.remote.RemoteHosts.uptime")
    def test_wait_reboot_since_uptime_fails(self, mocked_uptime, mocked_sleep, side_effect):
        """It should raise RemoteCheckError if unable to check the uptime on any host."""
        since = datetime.utcnow()
        mocked_uptime.side_effect = side_effect
        with pytest.raises(
            remote.RemoteCheckError,
            match=r"Unable to get uptime for host\[1-9\]",
        ):
            self.remote_hosts.wait_reboot_since(since)

        # wait_reboot_since() sets tries to 240 and dry_run is False.
        assert mocked_uptime.call_count == 240
        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    @mock.patch("spicerack.remote.RemoteHosts.uptime")
    def test_wait_reboot_since_uptime_too_big(self, mocked_uptime, mocked_sleep):
        """It should raise RemoteCheckError if any host doesn't have a small-enough uptime."""
        since = datetime.utcnow()
        mocked_uptime.return_value = [(NodeSet("host[1-9]"), 30.0)]
        with pytest.raises(
            remote.RemoteCheckError,
            match=r"Uptime for host\[1-9\] higher than threshold",
        ):
            self.remote_hosts.wait_reboot_since(since)

        mocked_uptime.assert_called_with(print_progress_bars=True)
        assert mocked_sleep.called

    def test_uptime_ok(self):
        """It should gather the current uptime from the target hosts."""
        nodes_a = "host1"
        nodes_b = "host[2-9]"
        mock_cumin(
            self.mocked_transports,
            0,
            retvals=[[(nodes_a, b"1514768400"), (nodes_b, b"1514768401")]],
        )
        uptimes = self.remote_hosts.uptime()
        assert sorted(uptimes) == sorted([(NodeSet(nodes_a), 1514768400.0), (NodeSet(nodes_b), 1514768401.0)])

    def test_uptime_invalid(self):
        """It should raise RemoteError if unable to parse the output as uptime."""
        mock_cumin(self.mocked_transports, 0, retvals=[[("host1", b"invalid")]])
        with pytest.raises(remote.RemoteError, match="Unable to extract data with <lambda> for host1"):
            self.remote_hosts.uptime()

    def test_results_to_list_callback(self):
        """It should return the output string coverted by the callback."""
        results = (item for item in [(self.hosts, MsgTreeElem(b"test", parent=MsgTreeElem()))])

        extracted = remote.RemoteHosts.results_to_list(results, callback=lambda x: x.upper())
        assert sorted(extracted) == sorted([(self.hosts, "TEST")])

    def test_results_to_list_no_callback(self):
        """It should return the output string without any conversion."""
        results = (item for item in [(self.hosts, MsgTreeElem(b"test", parent=MsgTreeElem()))])
        extracted = remote.RemoteHosts.results_to_list(results)
        assert sorted(extracted) == sorted([(self.hosts, "test")])

    def test_results_to_list_callback_raise(self):
        """It should raise RemoteError if the callback call raises any exception."""
        results = (item for item in [(self.hosts, MsgTreeElem(b"test", parent=MsgTreeElem()))])
        with pytest.raises(
            remote.RemoteError,
            match=r"Unable to extract data with <lambda> for host\[1-9\] from: test",
        ):
            remote.RemoteHosts.results_to_list(results, callback=lambda x: 1 / 0)

    def test_split_simple(self):
        """It should correctly split a simple remote."""
        results = list(self.remote_hosts.split(2))
        assert len(results) == 2
        assert len(results[0]) == 5
        assert len(results[1]) == 4

    def test_split_too_high(self):
        """It should correctly split the remote even if the slices are too many."""
        results = list(self.remote_hosts.split(15))
        assert len(results) == 9
        for result in results:
            assert len(result) == 1
            assert result._dry_run is False  # pylint: disable=protected-access

    @mock.patch("spicerack.remote.transport.Transport.new")
    def test_using_sudo_prepends_when_command_is_string(self, mocked_transport_new):
        """Test that using sudo prepends when command is string."""
        nodes_a = "host1"
        mock_cumin(self.mocked_transports, 0, retvals=[[(nodes_a, b"")]])
        mocked_worker = mock.MagicMock()
        mocked_worker.execute.return_value = 0
        mocked_transport_new.return_value = mocked_worker

        remote.RemoteHosts(self.config, NodeSet(nodes_a), dry_run=False, use_sudo=True).run_sync("command")

        assert mocked_worker.commands == ["sudo -i command"]

    @mock.patch("spicerack.remote.transport.Transport.new")
    def test_using_sudo_prepends_when_command_is_commands(self, mocked_transport_new):
        """Test that using sudo prepends when command is commands."""
        expected_command = remote.Command("sudo -i command", ok_codes=[0])
        nodes_a = "host1"
        mock_cumin(self.mocked_transports, 0, retvals=[[(nodes_a, b"")]])
        mocked_worker = mock.MagicMock()
        mocked_worker.execute.return_value = 0
        mocked_transport_new.return_value = mocked_worker

        remote.RemoteHosts(self.config, NodeSet(nodes_a), dry_run=False, use_sudo=True).run_sync(
            remote.Command("command")
        )

        assert mocked_worker.commands == [expected_command]

    @mock.patch("spicerack.remote.transport.Transport.new")
    def test_using_sudo_prepends_when_command_is_commands_and_str(self, mocked_transport_new):
        """Test that using sudo prepends when command is commands and str."""
        expected_command = remote.Command("sudo -i command", ok_codes=[0])
        nodes_a = "host1"
        mock_cumin(self.mocked_transports, 0, retvals=[[(nodes_a, b"")]])
        mocked_worker = mock.MagicMock()
        mocked_worker.execute.return_value = 0
        mocked_transport_new.return_value = mocked_worker

        remote.RemoteHosts(self.config, NodeSet(nodes_a), dry_run=False, use_sudo=True).run_sync(
            remote.Command("command"), "command"
        )

        assert mocked_worker.commands == [expected_command, "sudo -i command"]



"""ElasticsearchCluster module test."""
import itertools
from datetime import datetime, timedelta
from typing import Dict
from unittest import mock

import pytest
from cumin import NodeSet
from elasticsearch import ConflictError, Elasticsearch, RequestError, TransportError
from wmflib.config import load_yaml_config
from wmflib.prometheus import Prometheus

from spicerack import elasticsearch_cluster as ec
from spicerack.administrative import Reason
from spicerack.elasticsearch_cluster import NodesGroup
from spicerack.remote import Remote, RemoteHosts
from spicerack.tests import get_fixture_path

ELASTICSEARCH_CONFIG = load_yaml_config(get_fixture_path("elasticsearch", "config.yaml"))


def test_create_elasticsearch_clusters():
    """It should return an instance of ElasticsearchCluster."""
    target = ec.create_elasticsearch_clusters(ELASTICSEARCH_CONFIG, "search_eqiad", ["some_core_dc"], None, None)
    assert isinstance(target, ec.ElasticsearchClusters)


def test_create_elasticsearch_clusters_fail():
    """It should throw an ElasticsearchCluster Exception."""
    with pytest.raises(ec.ElasticsearchClusterError, match="No cluster group named search_test"):
        ec.create_elasticsearch_clusters(ELASTICSEARCH_CONFIG, "search_test", ["some_core_dc"], None, None)


def test_get_remote_hosts():
    """Test that RemoteHosts instance is returned."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.hosts = NodeSet("el[1-2]")
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    result = elastic_hosts.get_remote_hosts()
    assert isinstance(result, RemoteHosts)


def test_start_elasticsearch():
    """Test that start elasticsearch service is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.start_elasticsearch()
    mocked_remote_hosts.run_sync.assert_called_with("cat /etc/elasticsearch/instances | xargs systemctl start")


def test_stop_elasticsearch():
    """Test that stop elasticsearch service is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.stop_elasticsearch()
    mocked_remote_hosts.run_sync.assert_called_with("cat /etc/elasticsearch/instances | xargs systemctl stop")


def test_restart_elasticsearch():
    """Test that restart elasticsearch service is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.restart_elasticsearch()
    mocked_remote_hosts.run_sync.assert_called_with("cat /etc/elasticsearch/instances | xargs systemctl restart")


def test_depool_nodes():
    """Test that depool command is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.depool_nodes()
    mocked_remote_hosts.run_sync.assert_called_with("depool")


def test_pool_nodes():
    """Test that pool command is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.pool_nodes()
    mocked_remote_hosts.run_sync.assert_called_with("pool")


def test_wait_for_elasticsearch_up_delegates_to_all_nodes_groups():
    """Test that elasticsearch instance is called when wait for instance to come up."""
    node_group1 = mock.Mock(spec_set=NodesGroup)
    node_group1.check_all_nodes_up = mock.Mock()
    node_group2 = mock.Mock(spec_set=NodesGroup)
    node_group2.check_all_nodes_up = mock.Mock()
    remote_hosts = mock.Mock(spec_set=RemoteHosts)

    hosts = ec.ElasticsearchHosts(remote_hosts, [node_group1, node_group2], dry_run=False)
    hosts.wait_for_elasticsearch_up()

    assert node_group1.check_all_nodes_up.called
    assert node_group2.check_all_nodes_up.called


def test_wait_for_elasticsearch_does_no_check_when_in_dry_run():
    """In dry run we expect to always return that all nodes are up."""
    node_group = mock.Mock(spec_set=NodesGroup)
    node_group.check_all_nodes_up = mock.Mock()
    remote_hosts = mock.Mock(spec_set=RemoteHosts)

    hosts = ec.ElasticsearchHosts(remote_hosts, [node_group], dry_run=True)
    hosts.wait_for_elasticsearch_up()

    assert not node_group.check_all_nodes_up.called


@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_wait_for_elasticsearch_up_fails_if_one_node_is_down(mocked_sleep):
    """Test that elasticsearch instance is called when wait for instance to come up."""
    node_group1 = mock.Mock(spec_set=NodesGroup)
    node_group1.check_all_nodes_up = mock.Mock()
    node_group2 = mock.Mock(spec_set=NodesGroup)
    node_group2.check_all_nodes_up = mock.Mock(side_effect=ec.ElasticsearchClusterCheckError())
    remote_hosts = mock.Mock(spec_set=RemoteHosts)

    hosts = ec.ElasticsearchHosts(remote_hosts, [node_group1, node_group2], dry_run=False)

    with pytest.raises(ec.ElasticsearchClusterCheckError):
        hosts.wait_for_elasticsearch_up(timedelta(minutes=1))
    assert mocked_sleep.called


@mock.patch("wmflib.decorators.time.sleep", return_value=None)
def test_wait_for_elasticsearch_up_retries_on_failures(mocked_sleep):
    """Test that elasticsearch instance is called when wait for instance to come up."""
    node_group = mock.Mock(spec_set=NodesGroup)
    node_group.check_all_nodes_up = mock.Mock(side_effect=ec.ElasticsearchClusterCheckError())
    remote_hosts = mock.Mock(spec_set=RemoteHosts)

    hosts = ec.ElasticsearchHosts(remote_hosts, [node_group], dry_run=False)

    with pytest.raises(ec.ElasticsearchClusterCheckError):
        hosts.wait_for_elasticsearch_up(timedelta(minutes=1))
    assert node_group.check_all_nodes_up.call_count == 12
    assert mocked_sleep.called


def test_cluster_settings_are_unchanged_when_stopped_replication_is_dry_run():
    """Check that cluster routing in dry run mode is truly safe."""
    elasticsearch = Elasticsearch("endpoint:9200")
    elasticsearch.cluster.put_settings = mock.Mock(return_value=True)
    elasticsearch_cluster = ec.ElasticsearchCluster(elasticsearch, None, dry_run=True)
    with elasticsearch_cluster.stopped_replication():
        assert not elasticsearch.cluster.put_settings.called


def test_get_nodes_wraps_exceptions():
    """Get nodes should wrap exceptions from elasticsearch client."""
    elasticsearch = mock.Mock()
    elasticsearch.nodes.info = mock.Mock(side_effect=TransportError(500, "test"))
    remote = mock.Mock()

    cluster = ec.ElasticsearchCluster(elasticsearch, remote)
    with pytest.raises(ec.ElasticsearchClusterError):
        cluster.get_nodes()


class TestElasticsearchClusters:
    """Test class for Elasticsearch Clusters."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.elasticsearch1 = Elasticsearch("endpoint:9201")
        self.elasticsearch2 = Elasticsearch("endpoint:9202")
        self.cluster1 = ec.ElasticsearchCluster(self.elasticsearch1, None, dry_run=False)
        self.cluster2 = ec.ElasticsearchCluster(self.elasticsearch2, None, dry_run=False)
        self.clusters = [self.cluster1, self.cluster2]

    def default_elasticsearch_clusters(self):
        """Return simple default Elasticsearch clusters to DRY up test code."""
        return ec.ElasticsearchClusters(self.clusters, None, None, ["eqiad", "codfw"])

    def test_flush_markers_on_clusters(self):
        """Test that elasticsearch call to flush markers was properly made."""
        self.elasticsearch1.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch1.indices.flush_synced = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush_synced = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.flush_markers(timedelta(seconds=30))
        self.elasticsearch1.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch1.indices.flush_synced.assert_called_with(request_timeout=30)
        self.elasticsearch2.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch2.indices.flush_synced.assert_called_with(request_timeout=30)

    def test_flush_markers_on_clusters_fail(self):
        """Test that when conflict error is raised, execution continues."""
        self.elasticsearch1.indices.flush = mock.Mock(side_effect=ConflictError("test"))
        self.elasticsearch1.indices.flush_synced = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush_synced = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.flush_markers(timedelta(seconds=30))
        self.elasticsearch1.indices.flush_synced.assert_called_with(request_timeout=30)
        self.elasticsearch2.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch2.indices.flush_synced.assert_called_with(request_timeout=30)

    def test_flush_markers_on_clusters_fail_synced(self):
        """Test that when conflict error is raised during synced flush, execution continues."""
        self.elasticsearch1.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch1.indices.flush_synced = mock.Mock(side_effect=ConflictError("test"))
        self.elasticsearch2.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush_synced = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.flush_markers(timedelta(seconds=30))
        self.elasticsearch1.indices.flush_synced.assert_called_with(request_timeout=30)
        self.elasticsearch2.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch2.indices.flush_synced.assert_called_with(request_timeout=30)

    def test_when_all_shards_are_assigned_no_allocation_is_performed(self):
        """Test that shard allocation is not performed when all shards have been assigned on all clusters."""
        self.elasticsearch1.nodes.info = mock.Mock(return_value={"nodes": {"ELASTIC1": {"name": "el1-alpha"}}})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={"nodes": {"ELASTIC7": {"name": "el1-beta"}}})
        self.elasticsearch1.cat.shards = mock.Mock(
            return_value=[
                {"index": "index1", "shard": 2, "state": "ASSIGNED"},
                {"index": "index2", "shard": 4, "state": "ASSIGNED"},
            ]
        )
        self.elasticsearch2.cat.shards = mock.Mock(
            return_value=[
                {"index": "index3", "shard": 6, "state": "ASSIGNED"},
                {"index": "index4", "shard": 7, "state": "ASSIGNED"},
            ]
        )
        self.elasticsearch1.cluster.reroute = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.reroute = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.force_allocation_of_all_unassigned_shards()
        assert not self.elasticsearch1.cluster.reroute.called
        assert not self.elasticsearch2.cluster.reroute.called

    def test_force_allocation_of_all_unassigned_shards(self):
        """Test that elasticsearch performs cluster reroute with with unassigned shards on all clusters."""
        self.elasticsearch1.nodes.info = mock.Mock(return_value={"nodes": {"ELASTIC1": {"name": "el1-alpha"}}})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={"nodes": {"ELASTIC7": {"name": "el1-beta"}}})
        self.elasticsearch1.cat.shards = mock.Mock(
            return_value=[
                {"index": "index1", "shard": 2, "state": "UNASSIGNED"},
                {"index": "index2", "shard": 4, "state": "ASSIGNED"},
            ]
        )
        self.elasticsearch2.cat.shards = mock.Mock(
            return_value=[
                {"index": "index3", "shard": 6, "state": "ASSIGNED"},
                {"index": "index4", "shard": 7, "state": "UNASSIGNED"},
            ]
        )
        self.elasticsearch1.cluster.reroute = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.reroute = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.force_allocation_of_all_unassigned_shards()
        self.elasticsearch1.cluster.reroute.assert_called_with(
            retry_failed=True,
            body={
                "commands": [
                    {
                        "allocate_replica": {
                            "index": "index1",
                            "shard": 2,
                            "node": "el1-alpha",
                        }
                    }
                ]
            },
        )
        self.elasticsearch2.cluster.reroute.assert_called_with(
            retry_failed=True,
            body={
                "commands": [
                    {
                        "allocate_replica": {
                            "index": "index4",
                            "shard": 7,
                            "node": "el1-beta",
                        }
                    }
                ]
            },
        )

    def test_force_allocation_of_shards_with_failed_node(self):
        """Test that shard allocation command was called twice after it fails on the first node."""
        self.elasticsearch1.nodes.info = mock.Mock(
            return_value={
                "nodes": {
                    "ELASTIC1": {"name": "el1-alpha"},
                    "ELASTIC3": {"name": "el2-alpha"},
                }
            }
        )
        self.elasticsearch2.nodes.info = mock.Mock(
            return_value={
                "nodes": {
                    "ELASTIC7": {"name": "el1-beta"},
                    "ELASTIC4": {"name": "el2-beta"},
                }
            }
        )
        self.elasticsearch1.cat.shards = mock.Mock(
            return_value=[
                {"index": "index1", "shard": 2, "state": "UNASSIGNED"},
                {"index": "index2", "shard": 4, "state": "ASSIGNED"},
            ]
        )
        self.elasticsearch2.cat.shards = mock.Mock(
            return_value=[
                {"index": "index3", "shard": 6, "state": "ASSIGNED"},
                {"index": "index4", "shard": 7, "state": "UNASSIGNED"},
            ]
        )
        self.elasticsearch1.cluster.reroute = mock.Mock(side_effect=RequestError("test"))
        self.elasticsearch2.cluster.reroute = mock.Mock(side_effect=RequestError("test"))
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.force_allocation_of_all_unassigned_shards()
        assert self.elasticsearch1.cluster.reroute.call_count == 2
        assert self.elasticsearch2.cluster.reroute.call_count == 2

    def test_stopped_replication(self):
        """Check that context manager stops replication and then starts replication on each cluster."""
        self.elasticsearch1.cluster.put_settings = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.put_settings = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        with elasticsearch_clusters.stopped_replication():
            self.elasticsearch1.cluster.put_settings.assert_called_with(
                body={"transient": {"cluster.routing.allocation.enable": "primaries"}}
            )
            self.elasticsearch2.cluster.put_settings.assert_called_with(
                body={"transient": {"cluster.routing.allocation.enable": "primaries"}}
            )

        self.elasticsearch1.cluster.put_settings.assert_called_with(
            body={"transient": {"cluster.routing.allocation.enable": "all"}}
        )
        self.elasticsearch2.cluster.put_settings.assert_called_with(
            body={"transient": {"cluster.routing.allocation.enable": "all"}}
        )

    def test_frozen_writes_write_to_index(self):
        """Test that elasticsearch write to index is called to freeze writes."""
        self.elasticsearch1.index = mock.Mock(return_value=True)
        self.elasticsearch2.index = mock.Mock(return_value=True)
        self.elasticsearch1.delete = mock.Mock(return_value=True)
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        reason = Reason("test", "test_user", "test_host", task_id="T111222")
        cluster1 = ec.ElasticsearchCluster(self.elasticsearch1, None, dry_run=False)
        cluster2 = ec.ElasticsearchCluster(self.elasticsearch2, None, dry_run=False)
        elasticsearch_clusters = ec.ElasticsearchClusters([cluster1, cluster2], None, None, ["eqiad", "codfw"])
        with elasticsearch_clusters.frozen_writes(reason):
            assert self.elasticsearch1.index.called
            assert self.elasticsearch2.index.called

        assert self.elasticsearch1.delete.called
        assert self.elasticsearch2.delete.called

    def test_when_frozen_writes_fails_exception_is_raised(self):
        """Test that when elasticsearch write to index fails, an exception is raised.

        and a call to delete/unfreeze write is placed
        """
        self.elasticsearch1.index = mock.Mock(side_effect=TransportError(500, "test"))
        self.elasticsearch2.index = mock.Mock(return_value=True)
        self.elasticsearch1.delete = mock.Mock(return_value=True)
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        reason = Reason("test", "test_user", "test_host", task_id="T111222")
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        with pytest.raises(ec.ElasticsearchClusterError):
            with elasticsearch_clusters.frozen_writes(reason):
                assert self.elasticsearch1.index.called

            assert self.elasticsearch1.delete.called
            assert self.elasticsearch2.delete.called

    def test_when_unfreeze_writes_fails_exception_is_raised(self):
        """Test that when elasticsearch delete doc in index fails, an exception is raised.

        and a call to delete/unfreeze write is placed
        """
        self.elasticsearch1.index = mock.Mock(return_value=True)
        self.elasticsearch2.index = mock.Mock(return_value=True)
        self.elasticsearch1.delete = mock.Mock(side_effect=TransportError(500, "test"))
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        reason = Reason("test", "test_user", "test_host", task_id="T111222")
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        with pytest.raises(ec.ElasticsearchClusterError):
            with elasticsearch_clusters.frozen_writes(reason):
                assert self.elasticsearch1.index.called

            assert self.elasticsearch1.delete.called
            assert self.elasticsearch2.delete.called

    def test_no_call_to_freeze_write_in_dry_run(self):
        """Test that when dry run is enabled, call to write to cluster index to freeze write is not placed."""
        self.elasticsearch1.index = mock.Mock(return_value=True)
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        cluster1 = ec.ElasticsearchCluster(self.elasticsearch1, None, dry_run=True)
        cluster2 = ec.ElasticsearchCluster(self.elasticsearch2, None, dry_run=True)
        reason = Reason("test", "test_user", "test_host", task_id="T111222")
        elasticsearch_clusters = ec.ElasticsearchClusters([cluster1, cluster2], None, None, ["eqiad", "codfw"])
        with elasticsearch_clusters.frozen_writes(reason):
            assert not self.elasticsearch1.index.called
            assert not self.elasticsearch2.delete.called

    def test_wait_for_all_write_queues_with_queues_empty(self):
        """Ensure that we return None in the "happy path", when all queues are empty, meaning we didn't raise."""
        prometheus = mock.MagicMock(spec_set=Prometheus)
        prometheus.query = mock.MagicMock(
            return_value=[
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "0"],
                },
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "0"],
                },
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "0"],
                },
            ]
        )
        elasticsearch_clusters = ec.ElasticsearchClusters(self.clusters, None, prometheus, ["eqiad", "codfw"])

        assert elasticsearch_clusters.wait_for_all_write_queues_empty() is None

    def test_wait_for_all_write_queues_with_queues_non_empty(self):
        """Ensure that we raise an ElasticsearchClusterCheckError if write queues aren't empty."""
        prometheus = mock.MagicMock(spec_set=Prometheus)
        prometheus.query = mock.MagicMock(
            return_value=[
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "9535"],
                },
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "9608"],
                },
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "9536"],
                },
            ]
        )
        elasticsearch_clusters = ec.ElasticsearchClusters(self.clusters, None, prometheus, ["eqiad", "codfw"])

        with pytest.raises(ec.ElasticsearchClusterCheckError):
            elasticsearch_clusters.wait_for_all_write_queues_empty()

    def test_wait_for_all_write_queues_with_empty_response(self):
        """Ensure that we raise ElasticsearchClusterError if prometheus query fails to return any results."""
        prometheus = mock.MagicMock(spec_set=Prometheus)
        prometheus.query = mock.MagicMock(return_value=[])
        elasticsearch_clusters = ec.ElasticsearchClusters(self.clusters, None, prometheus, ["eqiad", "codfw"])

        with pytest.raises(ec.ElasticsearchClusterError):
            elasticsearch_clusters.wait_for_all_write_queues_empty()

    def test_write_queue_datacenters_get_queried(self):
        """Ensure that there is a prometheus query for each dc of write_queue_datacenters."""
        prometheus = mock.MagicMock(spec_set=Prometheus)
        prometheus.query = mock.MagicMock(
            return_value=[
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "0"],
                },
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "0"],
                },
                {
                    "metric": {"topic": "the_topic", "partition": "the_partition"},
                    "value": [1597958424.599, "0"],
                },
            ]
        )
        elasticsearch_clusters = ec.ElasticsearchClusters(self.clusters, None, prometheus, ["eqiad", "codfw"])
        elasticsearch_clusters.wait_for_all_write_queues_empty()

        assert prometheus.query.mock_calls == [
            mock.call(mock.ANY, "eqiad"),
            mock.call(mock.ANY, "codfw"),
        ]

    def test_wait_for_green_on_all_clusters_elastisearch_call(self):
        """Makes sure the call to elasticsearch.cluster.health is placed for each cluster."""
        self.elasticsearch1.cluster.health = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.health = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.wait_for_green(timedelta(seconds=13))
        assert self.elasticsearch1.cluster.health.called
        assert self.elasticsearch2.cluster.health.called

    @mock.patch("spicerack.elasticsearch_cluster.retry")
    def test_wait_for_green_correct_tries_test(self, retry):
        """Check that the number of tries is correctly computed."""
        self.elasticsearch1.cluster.health = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.health = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.wait_for_green(timedelta(seconds=20))
        assert retry.call_args[1]["tries"] == 2

    @mock.patch("spicerack.elasticsearch_cluster.retry")
    def test_wait_for_green_default_tries_test(self, retry):
        """Checks that a default value of 1 is returned when timeout is less than 10."""
        self.elasticsearch1.cluster.health = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.health = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.wait_for_green(timedelta(seconds=4))
        assert retry.call_args[1]["tries"] == 1

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_green_retry_test(self, mocked_sleep):
        """Test that the retry is called again when cluster health request throws an exception."""
        self.elasticsearch1.cluster.health = mock.Mock(side_effect=TransportError(500, "test"))
        self.elasticsearch2.cluster.health = mock.Mock(side_effect=TransportError(500, "test"))
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        with pytest.raises(ec.ElasticsearchClusterCheckError):
            elasticsearch_clusters.wait_for_green(timedelta(seconds=20))
            assert mocked_sleep.called
            assert self.elasticsearch1.cluster.health.call_count == 2
            assert self.elasticsearch2.cluster.health.call_count == 2

    def test_wait_for_yellow_w_no_moving_shards_on_all_clusters_elastisearch_call(self):
        """Makes sure the call to elasticsearch.cluster.health is placed for each cluster."""
        self.elasticsearch1.cluster.health = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.health = mock.Mock(return_value=True)
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        elasticsearch_clusters.wait_for_yellow_w_no_moving_shards(timedelta(seconds=13))
        assert self.elasticsearch1.cluster.health.called
        assert self.elasticsearch2.cluster.health.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    def test_wait_for_yellow_w_no_moving_shards_retry_test(self, mocked_sleep):
        """Test that the retry is called again when cluster health request throws an exception."""
        self.elasticsearch1.cluster.health = mock.Mock(side_effect=TransportError(500, "test"))
        self.elasticsearch2.cluster.health = mock.Mock(side_effect=TransportError(500, "test"))
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        with pytest.raises(ec.ElasticsearchClusterCheckError):
            elasticsearch_clusters.wait_for_yellow_w_no_moving_shards(timedelta(seconds=20))
            assert mocked_sleep.called
            assert self.elasticsearch1.cluster.health.call_count == 2
            assert self.elasticsearch2.cluster.health.call_count == 2

    def test_reset_read_only_is_sent_to_all_clusters(self):
        """Reset read only status should be sent to all clusters."""
        # This should really be an integration test but too much work to set up.
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        for client in [self.elasticsearch1, self.elasticsearch2]:
            client.indices.put_settings = mock.Mock()

        elasticsearch_clusters.reset_indices_to_read_write()

        for client in [self.elasticsearch1, self.elasticsearch2]:
            assert client.indices.put_settings.called

    def test_reset_read_only_wraps_exceptions(self):
        """Exceptions from underlying elasticsearch client should be wrapped."""
        elasticsearch_clusters = self.default_elasticsearch_clusters()
        for client in [self.elasticsearch1, self.elasticsearch2]:
            client.indices.put_settings = mock.Mock(side_effect=TransportError(500, "test"))

        with pytest.raises(ec.ElasticsearchClusterError):
            elasticsearch_clusters.reset_indices_to_read_write()
            assert not self.elasticsearch2.indices.put_settings.called


def test_get_next_clusters_nodes():
    """Test that next nodes belong in the same row on each cluster."""
    remote = mock.Mock(spec_set=Remote)
    since = datetime.utcfromtimestamp(20 / 1000)
    elasticsearch_clusters = ec.ElasticsearchClusters(
        mock_node_info(
            [
                {
                    "x1": json_node("elastic1001.example.com", "alpha", "row1", 10),
                    "x2": json_node("elastic1002.example.com", "alpha", "row1", 10),
                    "x3": json_node("elastic1003.example.com", "alpha", "row1", 30),
                },
                {
                    "y3": json_node("elastic1003.example.com", "gamma", "row1", 10),
                    "y4": json_node("elastic1004.example.com", "gamma", "row1", 10),
                },
            ]
        ),
        remote,
        None,
        ["eqiad", "codfw"],
    )
    elasticsearch_clusters.get_next_clusters_nodes(since, 4)
    nodes_not_restarted = remote.query.call_args[0][0]
    nodes_not_restarted = nodes_not_restarted.split(",")
    nodes_not_restarted.sort()
    assert nodes_not_restarted == [
        "elastic1001.example.com",
        "elastic1002.example.com",
        "elastic1003.example.com",
        "elastic1004.example.com",
    ]


def test_get_next_clusters_nodes_raises_error_when_size_is_less_than_one():
    """Test that next nodes belong in the same row on each cluster."""
    since = datetime.utcfromtimestamp(20 / 1000)
    elasticsearch_clusters = ec.ElasticsearchClusters(None, None, ["eqiad", "codfw"], None)
    with pytest.raises(ec.ElasticsearchClusterError):
        elasticsearch_clusters.get_next_clusters_nodes(since, 0)


def test_get_next_nodes_returns_less_nodes_than_specified():
    """Test that the nodes returned is less than specified based on if they have been restarted for each clusters."""
    remote = mock.Mock(spec_set=Remote)
    since = datetime.utcfromtimestamp(20 / 1000)
    elasticsearch_clusters = ec.ElasticsearchClusters(
        mock_node_info(
            [
                {
                    "x1": json_node("elastic1001.example.com", start_time=10),
                    "x2": json_node("elastic1002.example.com", start_time=10),
                    "x3": json_node("elastic1003.example.com", start_time=30),
                }
            ]
        ),
        remote,
        None,
        ["eqiad", "codfw"],
    )
    elasticsearch_clusters.get_next_clusters_nodes(since, 4)
    nodes_not_restarted = remote.query.call_args[0][0]
    nodes_not_restarted = nodes_not_restarted.split(",")
    nodes_not_restarted.sort()
    assert nodes_not_restarted == ["elastic1001.example.com", "elastic1002.example.com"]


def test_get_next_clusters_nodes_fails_with_masters_depending_on_each_other():
    """Tests that the restart fails in an obvious manner when nothing is restartable."""
    remote = mock.Mock(spec_set=Remote)
    since = datetime.utcfromtimestamp(20 / 1000)
    elasticsearch_clusters = ec.ElasticsearchClusters(
        mock_node_info(
            [
                {
                    "x1": json_node(
                        "elastic1001.example.com", cluster_name="alpha", start_time=10, master_capable=True
                    ),
                    "x2": json_node("elastic1002.example.com", cluster_name="alpha", start_time=10),
                },
                {
                    "x1": json_node("elastic1001.example.com", cluster_name="beta", start_time=10),
                    "x2": json_node("elastic1002.example.com", cluster_name="beta", start_time=10, master_capable=True),
                },
            ]
        ),
        remote,
        None,
        ["eqiad", "codfw"],
    )
    with pytest.raises(Exception) as excinfo:
        elasticsearch_clusters.get_next_clusters_nodes(since, 4)
    assert "graph has cycles" in str(excinfo.value)


def _eval_get_next_nodes(node_info, batch_size=4):
    def update_start(start_time, accept):
        for cluster in node_info:
            for node in cluster.values():
                if accept(node):
                    node["jvm"] = {"start_time_in_millis": start_time}

    update_start(0, lambda x: True)
    since = datetime.utcfromtimestamp(10 / 1000)
    for i in itertools.count(start=20, step=10):
        remote = mock.Mock(spec_set=Remote)
        elasticsearch_clusters = ec.ElasticsearchClusters(mock_node_info(node_info), remote, None, ["eqiad", "codfw"])
        res = elasticsearch_clusters.get_next_clusters_nodes(since, batch_size)
        if res is None:
            return
        # FIXME: We assert on the nodes that were queried via cumin and not on the return value of
        #  get_next_clusters_nodes(). This is a simplification to avoid mocking yet another return value.
        #  It is also a code smell. There is a refactoring waiting to happen to reduce / isolate the complexity
        #  of get_next_clusters_nodes().
        nodes_queried = remote.query.call_args[0][0]
        nodes_queried = nodes_queried.split(",")
        nodes_queried.sort()
        update_start(i, lambda x, nodes=nodes_queried: x["attributes"]["fqdn"] in nodes)
        yield nodes_queried


def test_get_next_nodes_returns_masters_from_separate_rows_one_at_a_time():
    """Test to verify masters ignore batching and start one at a time.

    Makes the bold assumption the cluster is properly deployed with only a single
    master capable node per row.
    """
    node_info = [
        {
            "m1": json_node("elastic1005.example.com", "alpha", "row2", 10, master_capable=True),
            "m2": json_node("elastic1006.example.com", "alpha", "row3", 10, master_capable=True),
        },
    ]

    seen = set()
    for batch in _eval_get_next_nodes(node_info):
        assert len(batch) == 1
        for node in batch:
            assert node not in seen
            seen.add(node)
    assert len(seen) == 2


def test_get_next_nodes_returns_masters_after_other_nodes():
    """Test to verify master nodes are restarted last.

    In this setup, mimicing prod, where some nodes are masters in two clusters,
    and some nodes are masters in one cluster, a3 and a4 must still be rebooted
    prior to a1 and a2, to ensure the alpha cluster masters are not restarted
    prior to their worker nodes.
    """
    node_info = [
        {
            "a1": json_node("elastic1001.example.com", "alpha", "row2", 10, master_capable=True),
            "a2": json_node("elastic1002.example.com", "alpha", "row3", 10, master_capable=True),
            "a3": json_node("elastic1003.example.com", "alpha", "row3", 10),
            "a4": json_node("elastic1004.example.com", "alpha", "row4", 10),
            "a5": json_node("elastic1005.example.com", "alpha", "row2", 10),
            "a6": json_node("elastic1006.example.com", "alpha", "row2", 10),
        },
        {
            "a1": json_node("elastic1001.example.com", "beta", "row2", 10, master_capable=True),
            "a3": json_node("elastic1003.example.com", "beta", "row3", 10, master_capable=True),
            "a5": json_node("elastic1005.example.com", "beta", "row2", 10),
        },
        {
            "a2": json_node("elastic1002.example.com", "delta", "row3", 10, master_capable=True),
            "a4": json_node("elastic1004.example.com", "delta", "row4", 10, master_capable=True),
            "a6": json_node("elastic1006.example.com", "delta", "row2", 10),
        },
    ]

    expect_batches = [
        {"elastic1005.example.com", "elastic1006.example.com"},
        {"elastic1003.example.com", "elastic1004.example.com"},
        {"elastic1001.example.com", "elastic1002.example.com"},
    ]
    expect_batch = None
    for batch in _eval_get_next_nodes(node_info):
        if not expect_batch:
            assert expect_batches, "unexpected batch of hosts returned"
            expect_batch = expect_batches.pop(0)
        unexpected = [node for node in batch if node not in expect_batch]
        assert not unexpected, "node not expected in current batch"
        expect_batch = expect_batch.difference(batch)
    assert not expect_batches, "expected batches of hosts remain"


def test_get_next_nodes_least_not_restarted():
    """Test to get rows that have the least not restarted nodes first on each cluster."""
    remote = mock.Mock(spec_set=Remote)
    since = datetime.utcfromtimestamp(20 / 1000)
    elasticsearch_clusters = ec.ElasticsearchClusters(
        mock_node_info(
            [
                {
                    "x4": json_node("elastic1005.example.com", "alpha", "row2", 10),
                    "x5": json_node("elastic1006.example.com", "alpha", "row2", 10),
                    "x6": json_node("elastic1007.example.com", "alpha", "row2", 30),
                },
                {
                    "x7": json_node("elastic1009.example.com", "beta", "row3", 10),
                    "x8": json_node("elastic1008.example.com", "beta", "row3", 50),
                    "x9": json_node("elastic1010.example.com", "beta", "row3", 30),
                },
            ]
        ),
        remote,
        None,
        ["eqiad", "codfw"],
    )
    elasticsearch_clusters.get_next_clusters_nodes(since, 4)
    # FIXME: We assert on the nodes that were queried via cumin and not on the return value of
    #  get_next_clusters_nodes(). This is a simplification to avoid mocking yet another return value.
    #  It is also a code smell. There is a refactoring waiting to happen to reduce / isolate the complexity
    #  of get_next_clusters_nodes().
    nodes_queried = remote.query.call_args[0][0]
    nodes_queried = nodes_queried.split(",")
    nodes_queried.sort()
    assert nodes_queried == ["elastic1009.example.com"]


def test_get_next_nodes_no_rows():
    """Test that all nodes have been restarted on all clusters."""
    since = datetime.utcfromtimestamp(20 / 1000)
    elasticsearch_clusters = ec.ElasticsearchClusters(
        mock_node_info(
            [
                {
                    "x3": json_node("elastic1003.example.com", "gamma", "row1", 30),
                    "x4": json_node("elastic1005.example.com", "gamma", "row2", 87),
                    "x5": json_node("elastic1006.example.com", "gamma", "row2", 77),
                },
                {
                    "x3": json_node("elastic1004.example.com", "alpha", "row1", 40),
                    "x4": json_node("elastic1005.example.com", "gamma", "row2", 89),
                    "x5": json_node("elastic1016.example.com", "alpha", "row2", 79),
                },
            ]
        ),
        None,
        None,
        ["eqiad", "codfw"],
    )
    result = elasticsearch_clusters.get_next_clusters_nodes(since, 2)
    assert result is None


def test_get_next_nodes_fails_when_rows_are_not_same():
    """Test that error is raised when clusters instances of the same node belong to different rows."""
    since = datetime.utcfromtimestamp(20 / 1000)
    elasticsearch_clusters = ec.ElasticsearchClusters(
        mock_node_info(
            [
                {
                    "x3": json_node("elastic1003.example.com", "gamma", "row1", 10),
                    "x4": json_node("elastic1005.example.com", "gamma", "row2", 87),
                },
                {
                    "x3": json_node("elastic1003.example.com", "alpha", "row6", 10),
                },
            ]
        ),
        None,
        None,
        ["eqiad", "codfw"],
    )
    with pytest.raises(AssertionError):
        elasticsearch_clusters.get_next_clusters_nodes(since, 2)


def test_nodes_group_aggregates_same_clusters():
    """Same cluster aggregated multiple times is ignored."""
    node1 = json_node("elastic1001.example.com", "alpha", row="row1")
    node2 = json_node("elastic1001.example.com", "alpha", row="row1")
    cluster = mock.Mock()
    group = NodesGroup(node1, cluster)
    group.accumulate(node2, cluster)

    assert len(group.clusters_instances) == 1


def test_nodes_group_fail_to_accumulate_with_different_fqdn():
    """Aggregation should only works if used on the same node."""
    node1 = json_node("elastic1001.example.com")
    node2 = json_node("elastic1002.example.com")
    cluster = mock.Mock()
    group = NodesGroup(node1, cluster)

    with pytest.raises(AssertionError):
        group.accumulate(node2, cluster)


def test_all_nodes_are_restarted():
    """All nodes are deemed restarted if they are all listed as cluster nodes."""
    node1 = json_node("elastic1001.example.com", cluster_name="alpha")
    node2 = json_node("elastic1001.example.com", cluster_name="beta")
    alpha, beta = mock_node_info([{"x1": node1}, {"y1": node2}])  # pylint: disable=unbalanced-tuple-unpacking

    group = NodesGroup(node1, alpha)
    group.accumulate(node2, beta)

    group.check_all_nodes_up()

    # whitebox testing for convenience
    # TODO: refactor mock_node_info() to expose the mock Elasticsearch instances
    assert alpha._elasticsearch.nodes.info.called  # pylint: disable=protected-access
    assert beta._elasticsearch.nodes.info.called  # pylint: disable=protected-access


def test_node_not_restarted():
    """Error is raised if a node is not listed as a member of the cluster."""
    node1 = json_node("elastic1001.example.com", cluster_name="alpha")
    node2 = json_node("elastic1001.example.com", cluster_name="beta")
    alpha, beta = mock_node_info([{"x1": node1}, {"y1": node2}])  # pylint: disable=unbalanced-tuple-unpacking

    group = NodesGroup(node1, alpha)
    group.accumulate(node2, beta)

    # remove elastic1001 from alpha cluster
    alpha._elasticsearch.nodes.info = mock.Mock(return_value={"nodes": {}})  # pylint: disable=protected-access

    with pytest.raises(ec.ElasticsearchClusterCheckError):
        group.check_all_nodes_up()


def json_node(
    fqdn: str,
    cluster_name: str = "alpha-cluster",
    row: str = "row1",
    start_time: int = 10,
    master_capable: bool = False,
) -> Dict:
    """Used to mock the elasticsearch node API."""
    hostname = fqdn.split(".", 1)[0]
    node_name = f"{hostname}-{cluster_name}"
    return {
        "name": node_name,
        "attributes": {
            "row": row,
            "hostname": hostname,
            "fqdn": fqdn,
        },
        "roles": ["master"] if master_capable else [],
        "settings": {
            "cluster": {
                "name": cluster_name,
            }
        },
        "jvm": {"start_time_in_millis": start_time},
    }


def mock_node_info(values):
    """Creates a list of ElasticsearchCluster which will return the given node info."""
    clusters = []
    port = 9200
    for nodes in values:
        elasticsearch = Elasticsearch(f"localhost:{port}")
        port += 1
        elasticsearch.nodes.info = mock.Mock(return_value={"nodes": nodes})
        cluster = ec.ElasticsearchCluster(elasticsearch, None, dry_run=False)
        clusters.append(cluster)
    return clusters



"""Icinga module."""
import json
import logging
import re
import shlex
import time
from contextlib import contextmanager
from datetime import timedelta
from typing import Dict, Iterator, List, Mapping, Optional, Sequence, Tuple, cast

from cumin import NodeSet
from cumin.transports import Command

from spicerack.administrative import Reason
from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError
from spicerack.remote import RemoteHosts
from spicerack.typing import TypeHosts

ICINGA_DOMAIN: str = "icinga.wikimedia.org"
MIN_DOWNTIME_SECONDS: int = 60  # Minimum time in seconds the downtime can be set
logger = logging.getLogger(__name__)


class CommandFile(str):
    """String class to represent an Icinga command file path with cache capabilities."""

    _command_files: Dict[Tuple[str, str], str] = {}  # Cache a command file per Icinga hostname and configuration file

    def __new__(cls, icinga_host: RemoteHosts, *, config_file: str = "/etc/icinga/icinga.cfg") -> "CommandFile":
        """Get the Icinga host command file where to write the commands and cache it.

        Arguments:
            icinga_host (spicerack.remote.RemoteHosts): the Icinga host instance.
            config_file (str, optional): the Icinga configuration file to check for the command file directive.

        Returns:
            str: the Icinga command file path on the Icinga host.

        Raises:
            spicerack.icinga.IcingaError: if unable to get the command file path.

        """
        # Can't use functools cache decorators because NodeSet are not hashable
        if len(icinga_host) != 1:
            raise IcingaError(f"Icinga host must match a single host, got: {icinga_host}")

        identifier = (str(icinga_host), config_file)

        if identifier in cls._command_files:
            return cast(CommandFile, cls._command_files[identifier])

        try:
            # Get the command_file value in the Icinga configuration.
            command = r"grep -P '\s*command_file\s*=.+' " + config_file
            command_file = ""
            for _, output in icinga_host.run_sync(
                command, is_safe=True, print_output=False, print_progress_bars=False
            ):  # Read only operation
                command_file = output.message().decode().split("=", 1)[-1].strip()

            if not command_file:
                raise ValueError(f"Empty or no value found for command_file configuration in {config_file}")

        except (SpicerackError, ValueError) as e:
            raise IcingaError(f"Unable to read command_file configuration in {config_file}") from e

        cls._command_files[identifier] = command_file
        return cast(CommandFile, command_file)


class IcingaError(SpicerackError):
    """Custom exception class for errors of this module."""


class IcingaCheckError(SpicerackCheckError):
    """Custom exception class for check errors of this module."""


class IcingaStatusParseError(IcingaError):
    """Custom exception class for errors while parsing the Icinga status."""


class IcingaStatusNotFoundError(IcingaError):
    """Custom exception class for a host missing from the Icinga status."""

    def __init__(self, hostnames: Sequence[str]):
        """Initializes an IcingaStatusNotFoundError instance.

        Arguments:
            hostnames (sequence): The hostnames not found in the Icinga status.

        """
        if len(hostnames) == 1:
            super().__init__(f"Host {hostnames[0]} was not found in Icinga status")
        else:
            hosts = ", ".join(hostnames)
            super().__init__(f"Hosts {hosts} were not found in Icinga status")


class HostsStatus(dict):
    """Represent the status of all Icinga checks for a set of hosts."""

    @property
    def optimal(self) -> bool:
        """Returns :py:data:`True` if all the hosts are optimal, :py:data:`False` otherwise.

        Returns:
            bool: whether all hosts are optimal.

        """
        return all(status.optimal for status in self.values())

    @property
    def non_optimal_hosts(self) -> List[str]:
        """Return the list of hostnames that are not in an optimal state.

        They can either not being up and running or have at least one failed service.

        Returns:
            list: a list of strings with the hostnames.

        """
        return [hostname for hostname, status in self.items() if not status.optimal]

    @property
    def failed_services(self) -> Dict[str, List[str]]:
        """Return the list of service names that are failing for each host that has at least one.

        Returns:
            dict: a dict with hostnames as keys and list of failing service name strings as values.

        """
        return {status.name: status.failed_services for status in self.values() if not status.optimal}

    @property
    def failed_hosts(self) -> List[str]:
        """Return the list of hostnames that are not up and running. They can either be down or unreachable.

        Returns:
            list: the list of strings with the hostnames.

        """
        return [status.name for status in self.values() if status.state != HostStatus.STATE_UP]


class HostStatus:
    """Represent the status of all Icinga checks for a single host."""

    STATE_UP = "UP"
    """:py:class:`str`: the Icinga value for a host that is up and running. The other values for the Icinga host state
    are ``DOWN`` and ``UNREACHABLE``."""

    def __init__(
        self,
        *,
        name: str,
        state: str,
        optimal: bool,
        downtimed: bool,
        notifications_enabled: bool,
        failed_services: Optional[Sequence[Mapping]] = None,
        services: Optional[Sequence[Mapping]] = None,
    ):
        """Initialize the instance.

        Either `services` or `failed_services` may be present, depending on the flags passed to icinga-status.

        Arguments:
            name (str): the hostname.
            state (str): the Icinga state for the host, one of ``UP``, ``DOWN``, UNREACHABLE``.
            optimal (bool): whether the host is in optimal state (all green).
            downtimed (bool): whether the host is currently downtimed.
            notifications_enabled: (bool): whether the host has notifications enabled.
            failed_services (list, optional): a list of dictionaries representing the failed services.
            services (list, optional): a list of dictionaries giving detailed service status.

        """
        self.name = name
        self.state = state
        self.optimal = optimal
        self.downtimed = downtimed
        self.notifications_enabled = notifications_enabled
        # TODO: could be improved creating a ServiceStatus class
        self.services = services if services is not None else []
        self.failed_services_raw = failed_services if failed_services is not None else []

    @property
    def failed_services(self) -> List[str]:
        """Return the list of service names that are failing.

        Returns:
            list: a list of strings with the check names.

        """
        return [service["name"] for service in self.failed_services_raw]


class IcingaHosts:
    """Class to manage the Icinga checks of a given set of hosts."""

    def __init__(self, icinga_host: RemoteHosts, target_hosts: TypeHosts, *, verbatim_hosts: bool = False) -> None:
        """Initialize the instance.

        Arguments:
            icinga_host (spicerack.remote.RemoteHosts): the RemoteHosts instance for the Icinga server.
            target_hosts (spicerack.typing.TypeHosts): the target hosts either as a NodeSet instance or a sequence of
                strings.
            verbatim_hosts (bool, optional): if :py:data:`True` use the hosts passed verbatim as is, if instead
                :py:data:`False`, the default, consider the given target hosts as FQDNs and extract their hostnames to
                be used in Icinga.

        """
        if not verbatim_hosts:
            target_hosts = [target_host.split(".")[0] for target_host in target_hosts]

        if isinstance(target_hosts, NodeSet):
            self._target_hosts = target_hosts
        else:
            self._target_hosts = NodeSet.fromlist(target_hosts)

        if not self._target_hosts:
            raise IcingaError("Got empty target hosts list.")

        self._command_file = CommandFile(icinga_host)  # This validates also that icinga_host matches a single server.
        self._icinga_host = icinga_host
        self._verbatim_hosts = verbatim_hosts

    @contextmanager
    def downtimed(
        self, reason: Reason, *, duration: timedelta = timedelta(hours=4), remove_on_error: bool = False
    ) -> Iterator[None]:
        """Context manager to perform actions while the hosts are downtimed on Icinga.

        Arguments:
            reason (spicerack.administrative.Reason): the reason to set for the downtime on the Icinga server.
            duration (datetime.timedelta, optional): the length of the downtime period.
            remove_on_error: should the downtime be removed even if an exception was raised.

        Yields:
            None: it just yields control to the caller once Icinga has been downtimed and deletes the downtime once
            getting back the control.

        """
        self.downtime(reason, duration=duration)
        try:
            yield
        except BaseException:
            if remove_on_error:
                self.remove_downtime()
            raise
        else:
            self.remove_downtime()

    def downtime(self, reason: Reason, *, duration: timedelta = timedelta(hours=4)) -> None:
        """Downtime hosts on the Icinga server for the given time with a message.

        Arguments:
            reason (spicerack.administrative.Reason): the reason to set for the downtime on the Icinga server.
            duration (datetime.timedelta, optional): the length of the downtime period.

        """
        duration_seconds = int(duration.total_seconds())
        if duration_seconds < MIN_DOWNTIME_SECONDS:
            raise IcingaError(f"Downtime duration must be at least 1 minute, got: {duration}")

        try:
            self.get_status()  # Ensure all hosts are known to Icinga, ignoring the return value.
        except IcingaStatusNotFoundError as e:
            raise IcingaError(f"{e} - no hosts have been downtimed.") from e

        logger.info(
            "Scheduling downtime on Icinga server %s for hosts: %s",
            self._icinga_host,
            self._target_hosts,
        )
        start_time = str(int(time.time()))
        end_time = str(int(time.time() + duration_seconds))
        # TODO: SCHEDULE_HOST_DOWNTIME may not be needed, since a quick look at the Icinga source code suggests that
        #  SCHEDULE_HOST_SVC_DOWNTIME also downtimes the host itself, not just the services. But if it does so, that's
        #  an undocumented extra feature. For now we're keeping this call for consistency with the older icinga-downtime
        #  script, even though it may be redundant, and in the future we can evaluate whether it's unnecessary.
        self.run_icinga_command(
            "SCHEDULE_HOST_DOWNTIME",
            start_time,
            end_time,
            "1",  # Start at the start_time and end at the end_time.
            "0",  # Not triggered by another downtime.
            str(duration_seconds),
            reason.owner,
            reason.reason,
        )
        self.run_icinga_command(
            "SCHEDULE_HOST_SVC_DOWNTIME",
            start_time,
            end_time,
            "1",  # Start at the start_time and end at the end_time.
            "0",  # Not triggered by another downtime.
            str(duration_seconds),
            reason.owner,
            reason.reason,
        )
        try:  # Best effort attempt to ensure the downtime was applied. See T309447.
            self.wait_for_downtimed()
        except IcingaCheckError as e:
            logger.warning(e)

    @retry(
        tries=12,
        delay=timedelta(seconds=10),
        backoff_mode="constant",
        exceptions=(IcingaCheckError,),
        failure_message="Unable to verify all hosts got downtimed",
    )
    def wait_for_downtimed(self) -> None:
        """Poll the Icinga status to verify that the hosts got effectively downtimed.

        Raises:
            spicerack.icinga.IcingaError: if unable to verify that all hosts got downtimed.

        """
        not_downtimed = [hostname for hostname, host_status in self.get_status().items() if not host_status.downtimed]
        if not_downtimed:
            raise IcingaCheckError(f"Some hosts are not yet downtimed: {not_downtimed}")

    @contextmanager
    def services_downtimed(
        self,
        service_re: str,
        reason: Reason,
        *,
        duration: timedelta = timedelta(hours=4),
        remove_on_error: bool = False,
    ) -> Iterator[None]:
        """Context manager to perform actions while services are downtimed on Icinga.

        Arguments:
            service_re (str): the regular expression matching service names to downtime.
            reason (spicerack.administrative.Reason): the reason to set for the downtime on the Icinga server.
            duration (datetime.timedelta, optional): the length of the downtime period.
            remove_on_error (bool, optional): should the downtime be removed even if an exception was raised.

        Yields:
            None: it just yields control to the caller once Icinga has been downtimed and deletes the downtime once
            getting back the control.

        """
        self.downtime_services(service_re, reason, duration=duration)
        try:
            yield
        except BaseException:
            if remove_on_error:
                self.remove_service_downtimes(service_re)
            raise
        else:
            self.remove_service_downtimes(service_re)

    def downtime_services(self, service_re: str, reason: Reason, *, duration: timedelta = timedelta(hours=4)) -> None:
        """Downtime services on the Icinga server for the given time with a message.

        If there are multiple target_hosts, the set of matching services may vary from host to host (e.g. because a
        hostname, DB section, or other unique fact is included in the service name) and downtime_services will downtime
        each service on the correct target_host. If some hosts happen to have no matching services, they will be safely
        skipped. But if *no* hosts have matching services, IcingaError is raised (because the regex is probably wrong).

        Arguments:
            service_re (str): the regular expression matching service names to downtime.
            reason (spicerack.administrative.Reason): the reason to set for the downtime on the Icinga server.
            duration (datetime.timedelta, optional): the length of the downtime period.

        Raises:
            re.error: if service_re is an invalid regular expression.
            IcingaError: if no services on any target host match the regular expression.

        """
        duration_seconds = int(duration.total_seconds())
        if duration_seconds < MIN_DOWNTIME_SECONDS:
            raise IcingaError(f"Downtime duration must be at least 1 minute, got: {duration}")

        try:
            # This also validates the regular expression syntax and ensures all hosts are known to Icinga.
            status = self.get_status(service_re)
        except IcingaStatusNotFoundError as e:
            raise IcingaError(f"{e} - no hosts have been downtimed.") from e

        unique_services = set()
        for host_status in status.values():
            for service in host_status.services:
                unique_services.add(service["name"])
        unique_service_count = len(unique_services)
        matched_host_count = sum(1 if host_status.services else 0 for host_status in status.values())

        if not unique_services:
            raise IcingaError(f'No services on {self._target_hosts} matched "{service_re}"')

        logger.info(
            'Scheduling downtime on Icinga server %s for services "%s" for host%s: %s '
            "(matched %d unique service name%s on %d host%s)",
            self._icinga_host,
            service_re,
            "" if len(self._target_hosts) == 1 else "s",
            self._target_hosts,
            unique_service_count,
            "" if unique_service_count == 1 else "s",
            matched_host_count,
            "" if matched_host_count == 1 else "s",
        )
        start_time = str(int(time.time()))
        end_time = str(int(time.time() + duration_seconds))
        # This doesn't use self.run_icinga_command because if the service names are different, we'll set different
        # downtimes (and therefore run different Icinga commands) for each target_host.
        commands = []
        for hostname, host_status in status.items():
            for service in host_status.services:
                logger.debug('Downtiming "%s" on %s', service["name"], hostname)
                commands.append(
                    self._get_command_string(
                        "SCHEDULE_SVC_DOWNTIME",
                        hostname,
                        service["name"],
                        start_time,
                        end_time,
                        "1",  # Start at the start_time and end at the end_time.
                        "0",  # Not triggered by another downtime.
                        str(duration_seconds),
                        reason.owner,
                        reason.reason,
                    )
                )
        self._icinga_host.run_sync(*commands, print_output=False, print_progress_bars=False)

    def recheck_all_services(self) -> None:
        """Force recheck of all services associated with a set of hosts."""
        self.run_icinga_command("SCHEDULE_FORCED_HOST_SVC_CHECKS", str(int(time.time())))

    def recheck_failed_services(self) -> None:
        """Force recheck of all failed associated with a set of hosts."""
        status = self.get_status()
        if status.optimal:
            return
        commands = [
            self._get_command_string("SCHEDULE_FORCED_SVC_CHECK", hostname, service_name, str(int(time.time())))
            for hostname, failed in status.failed_services.items()
            for service_name in failed
        ]
        self._icinga_host.run_sync(*commands, print_output=False, print_progress_bars=False)

    def remove_downtime(self) -> None:
        """Remove a downtime from a set of hosts."""
        self.run_icinga_command("DEL_DOWNTIME_BY_HOST_NAME")

    def remove_service_downtimes(self, service_re: str) -> None:
        """Remove downtimes for services from a set of hosts.

        If there are multiple target_hosts, this method has the same behavior as downtime_services. If any matching
        service is not downtimed, it's silently skipped. (If one or more services exist matching the regex, but none of
        them is downtimed, this method does nothing.)

        Arguments:
            service_re (str): the regular expression matching service names to un-downtime.

        Raises:
            re.error: if service_re is an invalid regular expression.
            IcingaError: if no services on any target host match the regular expression.

        """
        status = self.get_status(service_re)  # This also validates the regular expression syntax.
        if not any(host_status.services for host_status in status.values()):
            raise IcingaError(f'No services on {self._target_hosts} matched "{service_re}"')

        logger.info(
            'Removing downtime on Icinga server %s for services "%s" for hosts: %s',
            self._icinga_host,
            service_re,
            self._target_hosts,
        )

        commands = []
        for hostname, host_status in status.items():
            for service in host_status.services:
                if not service["status"]["scheduled_downtime_depth"]:  # Skip if not downtimed.
                    continue
                logger.debug('Removing downtime for "%s" on %s', service["name"], hostname)
                # DEL_DOWNTIME_BY_HOST_NAME is misleadingly named -- it also accepts an optional service name argument.
                commands.append(self._get_command_string("DEL_DOWNTIME_BY_HOST_NAME", hostname, service["name"]))
        if commands:
            self._icinga_host.run_sync(*commands, print_output=False, print_progress_bars=False)
        else:
            logger.info("No services downtimed, nothing to do.")

    def run_icinga_command(self, command: str, *args: str) -> None:
        """Execute an Icinga command on the Icinga server for all the current hosts.

        This lower level API is meant to be used when the higher level API exposed in this class does not cover a
        given use case. The arguments passed to the underlying Icinga command will be the hostname plus all the
        arguments passed to this method. Hence it can be used only with Icinga commands that require a hostname.
        See the link below for more details on the available Icinga commands and their arguments.

        Arguments:
            command (str): the Icinga command to execute.
            *args (str): optional positional arguments to pass to the command.

        See Also:
            https://icinga.com/docs/icinga1/latest/en/extcommands2.html

        """
        commands = [self._get_command_string(command, target_host, *args) for target_host in self._target_hosts]
        self._icinga_host.run_sync(*commands, print_output=False, print_progress_bars=False)

    def get_status(self, service_re: str = "") -> HostsStatus:
        """Get the current status of the given hosts from Icinga.

        Arguments:
            service_re (str): if non-empty, the regular expression matching service names

        Returns:
            spicerack.icinga.HostsStatus: the instance that represents the status for the given hosts.

        Raises:
            IcingaError: if unable to get the status.
            IcingaStatusParseError: when failing to parse the status.
            IcingaStatusNotFoundError: if a host is not found in the Icinga status.
            re.error: if service_re is an invalid regular expression.

        """
        if service_re:
            # Compile the regex and ignore the result, in order to raise re.error if it's malformed.
            re.compile(service_re)

        # icinga-status exits with non-zero exit code on missing and non-optimal hosts.
        verbatim = " --verbatim-hosts" if self._verbatim_hosts else ""
        services = (" --services " + shlex.quote(service_re)) if service_re else ""
        command = Command(
            f'/usr/local/bin/icinga-status -j{verbatim}{services} "{self._target_hosts}"',
            ok_codes=[],
        )
        for _, output in self._icinga_host.run_sync(
            command, is_safe=True, print_output=False, print_progress_bars=False
        ):  # icinga-status is a read-only script
            json_status = output.message().decode()
            break
        else:
            raise IcingaError("Unable to get the status for the given hosts, no output from icinga-status")

        try:
            status = json.loads(json_status)
        except json.JSONDecodeError as e:
            raise IcingaStatusParseError("Unable to parse Icinga status") from e

        missing_hosts = [hostname for hostname, host_status in status.items() if host_status is None]
        if missing_hosts:
            raise IcingaStatusNotFoundError(missing_hosts)

        return HostsStatus({hostname: HostStatus(**host_status) for hostname, host_status in status.items()})

    def wait_for_optimal(self) -> None:
        """Waits for an icinga optimal status, else raises an exception.

        This function will first instruct icinga to recheck all failed services
        and then wait until all services are in an optimal status.  If an
        optimal status is not reached in 6 minutes then we raise IcingaError

        Raises:
            IcingaError: if the status is not optimal.

        """

        @retry(
            tries=15,
            delay=timedelta(seconds=3),
            backoff_mode="linear",
            exceptions=(IcingaError,),
        )
        def check() -> None:
            status = self.get_status()
            if not status.optimal:
                failed = [f"{k}:{','.join(v)}" for k, v in status.failed_services.items()]
                raise IcingaError("Not all services are recovered: " + " ".join(failed))

        self.recheck_failed_services()
        check()

    def _get_command_string(self, *args: str) -> str:
        """Get the Icinga command to execute given the current arguments.

        Arguments:
            *args (str): positional arguments to use to compose the Icinga command string.

        Returns:
            str: the command line to execute on the Icinga host.

        """
        args_str = ";".join(args)
        bash_cmd = f'echo -n "[{int(time.time())}] {args_str}" > {self._command_file} '
        return "bash -c " + shlex.quote(bash_cmd)



"""Interactive module tests."""
import logging
import uuid
from pathlib import Path
from unittest import mock

from spicerack import _log as log

GENERIC_LOG_RECORD = logging.LogRecord("module", logging.DEBUG, "/source/file.py", 1, "message", [], None)
CUMIN_LOG_RECORD = logging.LogRecord("cumin.module", logging.DEBUG, "/cumin/source/file.py", 1, "message", [], None)
logger = logging.getLogger(__name__)


def _assert_match_in_tmpdir(match, tmp_dir):
    """Given a match string, assert that it's present in all files in tmp_dir."""
    tmp_dir = Path(tmp_dir)  # Newer versions pass a LocalPath, older a string.
    for logfile in tmp_dir.iterdir():
        with open(tmp_dir / logfile, "r") as f:
            assert match in f.read()


def _reset_logging_module():
    """Reset the logging module removing all handlers and filters."""
    for log_logger in (log.root_logger, log.irc_logger):
        list(map(log_logger.removeHandler, log_logger.handlers))
        list(map(log_logger.removeFilter, log_logger.filters))


def test_cumin_filter_pass():
    """The FilterOutCumin filter() method should let a normal log record pass."""
    log_filter = log.FilterOutCumin()
    ret = log_filter.filter(GENERIC_LOG_RECORD)
    assert ret == 1


def test_cumin_filter_blocks_cumin():
    """The FilterOutCumin filter() method should block a Cumin's log record."""
    log_filter = log.FilterOutCumin()
    ret = log_filter.filter(CUMIN_LOG_RECORD)
    assert ret == 0


def test_setup_logging_no_irc(tmpdir, caplog):
    """Calling setup_logging() should setup all the handlers of the root logger."""
    log.setup_logging(Path(tmpdir.strpath), "task", "user")
    message = str(uuid.uuid4())
    logger.info(message)
    assert message in caplog.text
    _assert_match_in_tmpdir(message, tmpdir.strpath)
    _reset_logging_module()


@mock.patch("wmflib.irc.socket")
def test_setup_logging_with_irc(mocked_socket, tmpdir, caplog):
    """Calling setup_logging() with host and port should also setup the IRC logger."""
    log.setup_logging(Path(tmpdir.strpath), "task", "user", host="host", port=123, dry_run=False)
    message = str(uuid.uuid4())
    log.irc_logger.info(message)

    assert mock.call.socket().connect(("host", 123)) in mocked_socket.mock_calls
    assert message in caplog.text
    _assert_match_in_tmpdir(message, tmpdir.strpath)
    _reset_logging_module()


def test_setup_logging_dry_run(capsys, tmpdir, caplog):
    """Calling setup_logging() when in dry run mode should setup all the handlers and the stdout with DEBUG level."""
    log.setup_logging(Path(tmpdir.strpath), "task", "user", dry_run=True)
    message = str(uuid.uuid4())
    logger.info(message)

    _, err = capsys.readouterr()
    assert message in err
    assert "DRY-RUN" in err
    assert message in caplog.text
    _assert_match_in_tmpdir(message, tmpdir.strpath)
    _assert_match_in_tmpdir("DRY-RUN", tmpdir.strpath)
    _reset_logging_module()


def test_log_task_start(capsys, tmpdir, caplog):
    """Calling log_task_start() should log a START message for the task to both loggers."""
    log.setup_logging(Path(tmpdir.strpath), "task", "user")
    message = str(uuid.uuid4())
    log.log_task_start(message)

    logged_message = "START - " + message
    _, err = capsys.readouterr()
    assert logged_message in err
    assert logged_message in caplog.text
    _assert_match_in_tmpdir(logged_message, tmpdir.strpath)
    _reset_logging_module()


def test_log_task_start_dry_run(capsys, tmpdir, caplog):
    """Calling log_task_start() in dry-run mode should not print a START message for the task to the IRC logger."""
    log.setup_logging(Path(tmpdir.strpath), "task", "user", dry_run=True)
    message = str(uuid.uuid4())
    log.log_task_start(message)

    logged_message = "START - " + message
    _, err = capsys.readouterr()
    assert logged_message in err
    assert logged_message in caplog.text
    _assert_match_in_tmpdir(logged_message, tmpdir.strpath)
    _assert_match_in_tmpdir("DRY-RUN", tmpdir.strpath)
    _reset_logging_module()


def test_log_task_end(capsys, tmpdir, caplog):
    """Calling log_task_end() should print an END message for the task."""
    log.setup_logging(Path(tmpdir.strpath), "task", "user", dry_run=False)
    message = str(uuid.uuid4())
    log.log_task_end("success", message)

    logged_message = "END (success) - " + message
    _, err = capsys.readouterr()
    assert logged_message in err
    assert logged_message in caplog.text
    _assert_match_in_tmpdir(logged_message, tmpdir.strpath)
    _reset_logging_module()



"""Multiple cookbook classes."""
from spicerack.cookbook import CookbookBase, CookbookRunnerBase


class CookbookA(CookbookBase):
    """Multiple cookbook classes."""

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner()


class CookbookB(CookbookBase):
    """Multiple cookbook classes."""

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner()


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    def run(self):
        """As required by the parent class."""



"""Service module."""
import logging
from collections import abc
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import timedelta
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Dict, Iterator, List, Optional, Sequence, Union

from spicerack.administrative import Reason
from spicerack.alertmanager import AlertmanagerHosts, MatchersType
from spicerack.confctl import ConftoolEntity
from spicerack.dnsdisc import Discovery
from spicerack.exceptions import SpicerackError
from spicerack.remote import Remote

logger = logging.getLogger(__name__)


class ServiceError(SpicerackError):
    """Generic exception class for errors in the service module."""


class ServiceNotFoundError(ServiceError):
    """Exception class raised when a service is not found."""


class TooManyDiscoveryRecordsError(ServiceError):
    """Exception class raised when more than one DNS Discovery record is present but not name was specified."""


class DiscoveryRecordNotFoundError(ServiceError):
    """Exception class raised when a DNS Discovery record is not found by name or there is none."""


@dataclass(frozen=True)
class ServiceDiscoveryRecord:
    """Represents the DNS Discovery attributes of the service.

    See Also:
        Puppet's ``modules/wmflib/types/service/discovery.pp``.

    Arguments:
        active_active (bool): :py:data:`True` if the service is active/active in DNS Discovery.
        dnsdisc (str): the name used in conftool for the discovery record.
        instance (spicerack.dnsdisc.Discovery): the DNS Discovery intance to operate on the service.

    """

    active_active: bool
    dnsdisc: str
    instance: Discovery


class ServiceDiscovery(abc.Iterable):
    """Represents the service Discovery records collection as list-like object with helper methods.

    The discovery behaves like an Iterator, so it can be iterated in list-comprehension and similar contructs.
    It supports also ``len()``, to quickly know how many records are configured for the service.

    """

    def __init__(self, records: Sequence[ServiceDiscoveryRecord]):
        """Initialize the instance with the records.

        Arguments:
            records (iterable, optional): the DNS Discovery records.

        """
        self.records = records

    def __iter__(self) -> Iterator[ServiceDiscoveryRecord]:
        """Iterate over the DNS Discovery records in the instance.

        Returns:
            iterator: iterator over the records.

        """
        return iter(self.records)

    def __len__(self) -> int:
        """Return the length of the instance.

        Returns:
            int: the number of DNS Discovery records.

        """
        return len(self.records)

    def depool(self, site: str, *, name: str = "") -> None:
        """Depool the service from the given site in DNS Discovery.

        Args:
            site (str): the datacenter to depool the service from.
            name (str, optional): the dnsdisc name of the DNS Discovery record to depool. If empty, and there is only
                one record, it will depool that one.

        Raises:
            spicerack.service.DiscoveryRecordNotFoundError: if there are no records at all or the record with the given
            name can't be found.
            spicerack.service.TooManyDiscoveryRecordsError: if the name is an empty string and there is more than one
            record.

        """
        self.get(name).instance.depool(site)

    def pool(self, site: str, *, name: str = "") -> None:
        """Pool the service to the given site in DNS Discovery.

        Args:
            site (str): the datacenter to pool the service to.
            name (str, optional): the dnsdisc name of the DNS Discovery record to pool. If empty, and there is only
                one record, it will pool that one.

        Raises:
            spicerack.service.DiscoveryRecordNotFoundError: if there are no records at all or the record with the given
            name can't be found.
            spicerack.service.TooManyDiscoveryRecordsError: if the name is an empty string and there is more than one
            record.

        """
        self.get(name).instance.pool(site)

    @contextmanager
    def depooled(self, site: str, *, name: str = "", repool_on_error: bool = False) -> Iterator[None]:
        """Context manager to act while the service is depooled from the given site in DNS Discovery.

        It will not repool the service on the given site if any exception is raised within the context manager context,
        unless ``repool_on_error`` is set to :py:data:`True`.

        Args:
            site (str): the datacenter to depool the service from.
            name (str, optional): the dnsdisc name of the DNS Discovery record to depool. If empty, and there is only
                one record, it will depool that one.
            repool_on_error (bool, optional): whether to repool the site on error or not.

        Yields:
            None

        Raises:
            spicerack.service.DiscoveryRecordNotFoundError: if there are no records at all or the record with the given
            name can't be found.
            spicerack.service.TooManyDiscoveryRecordsError: if the name is an empty string and there is more than one
            record.

        """
        self.depool(site, name=name)
        try:
            yield
        except Exception:
            if repool_on_error:
                self.pool(site, name=name)
            raise
        else:
            self.pool(site, name=name)

    def get(self, name: str = "") -> ServiceDiscoveryRecord:
        """Return the DNS Discovery record for the given name, raise an exception if not found.

        Arguments:
            name (str, optional): the dnsdic name of the DNS Discovery record. If set to an empty string, and the
                service has only one service, it will return that one.

        Raises:
            spicerack.service.DiscoveryRecordNotFoundError: if there are no records at all or the record with the given
            name can't be found.
            spicerack.service.TooManyDiscoveryRecordsError: if the name is an empty string and there is more than one
            record.

        """
        if not self.records:
            raise DiscoveryRecordNotFoundError("No DNS Discovery record present.")

        if name:
            matching = [record for record in self.records if record.dnsdisc == name]
            if not matching:
                raise DiscoveryRecordNotFoundError(f"Unable to find DNS Discovery record {name}.")
            if len(matching) > 1:
                raise TooManyDiscoveryRecordsError(
                    f"There are {len(matching)} DNS Discovery records matching name {name}."
                )
            return matching[0]

        if len(self.records) > 1:
            raise TooManyDiscoveryRecordsError(
                f"There are {len(self.records)} DNS Discovery records but dnsdisc was not set."
            )
        return self.records[0]


@dataclass(frozen=True)
class ServiceIPs:
    """Represent the service IPs.

    See Also:
        Puppet's ``modules/wmflib/types/service/ipblock.pp``.

    Arguments:
        data (dict): a dictionary representing the service IPs data as defined in ``service::catalog``.

    """

    data: Dict[str, Dict[str, str]]

    @property
    def all(self) -> List[Union[IPv4Address, IPv6Address]]:
        """Return all service IPs.

        Returns:
            list: a list of IP addresses.

        """
        return [ip_address(j) for i in self.data.values() for j in i.values()]

    @property
    def sites(self) -> List[str]:
        """Returns all the datacenters where there is at least one IP for the service.

        Returns:
            list: a list of strings.

        """
        return list(self.data.keys())

    def get(self, site: str, label: str = "default") -> Union[IPv4Address, IPv6Address, None]:
        """Get the IP for a given datacenter and optional label.

        Arguments:
            site (str): the datacenter to filter for.
            label (str, optional): the label of the IP.

        Returns:
            ipaddress.IPv4Address: if the matched IP is an IPv4.
            ipaddress.IPv6Address: if the matched IP is an IPv6.
            None: if there is no IP matching the criteria.

        """
        ip_str = self.data.get(site, {}).get(label, "")
        if ip_str:
            return ip_address(ip_str)

        return None


@dataclass(frozen=True)
class ServiceLVSConftool:
    """Represent the conftool configuration for the service for the load balancers.

    Arguments:
        cluster (str): name of the cluster tag in conftool.
        service (str): name of the service tag in conftool.

    """

    cluster: str
    service: str


@dataclass(frozen=True)
class ServiceLVS:
    """Represent the load balancer configuration for the service.

    See Also:
        Puppet's ``modules/wmflib/types/service/lvs.pp``.

    Arguments:
        conftool (spicerack.service.ServiceLVSConftool): the conftool configuration for the service.
        depool_threshold (str): the percentage of the cluster that Pybal will keep pooled anyway on failure.
        enabled (bool): whether the service is enabled on the load balancers.
        lvs_class (str): the traffic class of the service (e.g. ``low-traffic``).
        monitors (dict): a dictionary of Pybal monitors configured for the service.
        bgp (bool, optional): whether Pybal advertise the service via BGP or not.
        protocol (str, optional): the Internet protocol of the service.
        scheduler (str, optional): the IPVS scheduler used for the service.

    """

    conftool: ServiceLVSConftool
    depool_threshold: str
    enabled: bool
    lvs_class: str
    monitors: Dict[str, Dict]
    bgp: bool = True  # Default value in Puppet.
    protocol: str = "tcp"  # Default value in Puppet.
    scheduler: str = "wrr"  # Default value in Puppet.


@dataclass(frozen=True)
class ServiceMonitoringHostnames:
    """Represent the Icinga hostnames or FQDNs used to monitor the service in each datacenter.

    See Also:
        Puppet's ``modules/wmflib/types/service/monitoring.pp``.

    Arguments:
        data (dict): the service monitoring dictionary.

    """

    data: Dict[str, Dict[str, str]]

    @property
    def all(self) -> List[str]:
        """Return all service monitoring Icinga hostnames/FQDNs.

        Returns:
            list: the list of Icinga hostnames/FQDNs.

        """
        return [j for i in self.data.values() for j in i.values()]

    @property
    def sites(self) -> List[str]:
        """Return all the datacenters in which the monitoring is configured for.

        Returns:
            list: the list of datacenters.

        """
        return list(self.data.keys())

    def get(self, site: str) -> str:
        """Get the monitoring hostname/FQDN for the service in the given datacenter.

        Returns:
            str: the hostname/FQDN for the given datacenter if configured, empty string otherwise.

        """
        return self.data.get(site, {}).get("hostname", "")


@dataclass(frozen=True)
class ServiceMonitoring:
    """Represent the monitoring configuration for the service.

    See Also:
        Puppet's ``modules/wmflib/types/service/monitoring.pp``.

    Arguments:
        check_command (str): the Icinga check command used to monitor the service.
        sites (spicerack.service.ServiceMonitoringHostnames): the FQDNs used to monitor the service in each datacenter.
        contact_group (str, optional): the name of the Icinga contact group used for the service.
        notes_url (str, optional): the Icinga notes URL pointing to the service runbook.

    """

    check_command: str
    sites: ServiceMonitoringHostnames
    contact_group: str = ""
    notes_url: str = ""


@dataclass(frozen=True)
class Service:  # pylint: disable=too-many-instance-attributes
    """Class to represent a service as defined in Puppet's ``service::catalog`` hieradata.

    See Also:
        Puppet's ``modules/wmflib/types/service.pp``.

    Arguments:
        name (str): the service name.
        description (str): the service description.
        encryption (bool): whether TLS encryption is enabled or not on the service.
        ip (spicerack.service.ServiceIPs): the instance that represents all the service IPs.
        port (int): the port the service listen on.
        sites (list): the list of datacenters where the service is configured.
        state (str): the production state of the service (e.g. ``lvs_setup``).
        _alertmanager (spicerack.alertmanager:AlertmanagerHosts): the AlertmanagerHosts instance to perform downtime.
        aliases (list, optional): a list of aliases names for the service.
        discovery (spicerack.service.ServiceDiscovery, optional): the collection of
            :py:class:`spicerack.service.ServiceDiscoveryRecord` instances reprensenting the DNS Discovery capabilities
            of the service.
        lvs (spicerack.service.ServiceLVS, optional): the load balancer configuration.
        monitoring (spicerack.service.ServiceMonitoring, optional): the service monitoring configuration.
        page (bool, optional): whether the monitoring for this service does page or not.
        probes (list, optional): a list of probe dictionaries with all the parameters necessary to define the probes
            for this service.
        public_aliases (list, optional): the list of public aliases set for this service.
        public_endpoint (str, optional): the name of the public endpoint if present, empty string otherwise.
        role (str, optional): the service role name in Puppet if present, empty string otherwise.

    """

    name: str
    description: str
    encryption: bool
    ip: ServiceIPs  # pylint: disable=invalid-name
    port: int
    sites: List[str]
    state: str
    _alertmanager: AlertmanagerHosts
    aliases: List[str] = field(default_factory=list)
    discovery: Optional[ServiceDiscovery] = None
    lvs: Optional[ServiceLVS] = None
    monitoring: Optional[ServiceMonitoring] = None
    page: bool = True  # Default value in Puppet.
    probes: List[Dict] = field(default_factory=list)
    public_aliases: List[str] = field(default_factory=list)
    public_endpoint: str = ""
    role: str = ""

    def downtime(self, site: str, reason: Reason, *, duration: timedelta = timedelta(hours=4)) -> str:
        """Downtime the service on the given site in Alertmanager.

        Args:
            site (str): the datacenter where to silence the service.
            reason (spicerack.administrative.Reason): the silence reason.
            duration (datetime.timedelta, optional): how long to silence for.

        Raises:
            spicerack.service.ServiceError: if the service is not present in the given datacenter.

        Returns:
            str: the downtime ID.

        """
        return self._alertmanager.downtime(reason, matchers=self._get_downtime_matchers(site), duration=duration)

    @contextmanager
    def downtimed(
        self, site: str, reason: Reason, *, duration: timedelta = timedelta(hours=4), remove_on_error: bool = False
    ) -> Iterator[None]:
        """Context manager to perform actions while the service is downtimed in the given site in Alertmanager.

        Args:
            site (str): the datacenter where to silence the service.
            reason (spicerack.administrative.Reason): the silence reason.
            duration (datetime.timedelta, optional): how long to silence for.
            remove_on_error (bool, optional): should the downtime be removed even if an exception was raised.

        Raises:
            spicerack.service.ServiceError: if the service is not present in the given datacenter.

        """
        with self._alertmanager.downtimed(
            reason, matchers=self._get_downtime_matchers(site), duration=duration, remove_on_error=remove_on_error
        ):
            yield

    def _get_downtime_matchers(self, site: str) -> MatchersType:
        """Get the downtime matchers to use to downtime the service in Alertmanager.

        Args:
            site (str): the datacenter where to silence the service.

        """
        if site not in self.sites:
            raise ServiceError(f"Service {self.name} is not present in site {site}. Available sites are {self.sites}")

        return [
            {"name": "site", "value": site, "isRegex": False},
            {"name": "job", "value": r"^probes/.*", "isRegex": True},
        ]


class Catalog:
    """Class to represent the service catalog of Puppet's hierdata ``service::catalog``.

    The catalog behaves like an Iterator, so it can be iterated in list-comprehension and similar contructs.
    It supports also ``len()``, to quickly know how many services are loaded in the catalog.

    Examples:
        Get all service that are present in a given datacenter::

            >>> esams = [service for service in catalog if "esams" in service.sites]

        See how many services are configured::

            >>> num_services = len(catalog)

    """

    def __init__(self, catalog: Dict, confctl: ConftoolEntity, remote: Remote, dry_run: bool = True):
        """Initialize the instance.

        Args:
            catalog (dict): the content of Puppet's ``hieradata/common/service.yaml``.
            confctl (spicerack.confctl.ConftoolEntity): the instance to interact with confctl.
            remote (spicerack.remote.Remote): the instance to execute remote commands.
            dry_run (bool, optional): whether this is a DRY-RUN.

        """
        self._catalog = catalog
        self._confctl = confctl
        self._remote = remote
        self._dry_run = dry_run

    def __iter__(self) -> Iterator[Service]:
        """Iterate over the catalog services.

        Returns:
            iterator: iterator over the services.

        """
        return (self.get(name) for name in self._catalog)

    def __len__(self) -> int:
        """Return the length of the instance.

        Returns:
            int: the number of services in the catalog.

        """
        return len(self._catalog)

    @property
    def service_names(self) -> List[str]:
        """All defined service names.

        Returns:
            list: the list of service names.

        """
        return list(self._catalog.keys())

    def get(self, name: str) -> Service:
        """Get a single service by name.

        Examples:
            Get a specific service::

                >>> service = catalog.get("service_name")

        Arguments:
            name (str): the service name.

        Raises:
            spicerack.service.ServiceNotFoundError: if the service is not found.

        Returns:
            spicerack.service.Service: the instance of the matched service.

        """
        if name not in self._catalog:
            raise ServiceNotFoundError(f"Service {name} was not found in service::catalog")

        params = deepcopy(self._catalog[name])
        params["name"] = name
        params["ip"] = ServiceIPs(data=params["ip"])
        params["_alertmanager"] = AlertmanagerHosts(
            [f"{name}:{params['port']}"], verbatim_hosts=True, dry_run=self._dry_run
        )
        if "discovery" in params:
            discovery = []
            for disc in params["discovery"]:
                instance = Discovery(self._confctl, self._remote, [disc["dnsdisc"]], dry_run=self._dry_run)
                discovery.append(ServiceDiscoveryRecord(instance=instance, **disc))
            params["discovery"] = ServiceDiscovery(discovery)

        if "monitoring" in params:
            params["monitoring"]["sites"] = ServiceMonitoringHostnames(data=params["monitoring"]["sites"])
            params["monitoring"] = ServiceMonitoring(**params["monitoring"])

        if "lvs" in params:
            params["lvs"]["lvs_class"] = params["lvs"].pop("class")  # Rename class, reserved word in Python.
            params["lvs"]["conftool"] = ServiceLVSConftool(**params["lvs"]["conftool"])
            params["lvs"] = ServiceLVS(**params["lvs"])

        return Service(**params)



"""Constants module."""

CORE_DATACENTERS = ("eqiad", "codfw")
"""tuple: list of core datacenters."""

PUPPET_CA_PATH = "/etc/ssl/certs/Puppet_Internal_CA.pem"
"""str: the path to the Puppet Signing CA cert"""

KEYHOLDER_SOCK = "/run/keyholder/proxy.sock"
"""str: The path to the keyholder agent sock file used for cumin remote commands"""



"""Wrapper around etcdctl handling parameters and such."""
import logging
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, Optional, Tuple, Union, cast

from spicerack.exceptions import SpicerackError
from spicerack.remote import RemoteHosts, RemoteHostsAdapter

logger = logging.getLogger(__name__)
SimpleType = Union[str, int, bool]


class HealthStatus(Enum):
    """Health status."""

    # TODO: rename the following with UPPERCASE names
    healthy = auto()  # pylint: disable=invalid-name
    unhealthy = auto()  # pylint: disable=invalid-name


@dataclass(frozen=True)
class EtcdClusterHealthStatus:
    """Etcd cluster health status."""

    global_status: HealthStatus
    members_status: Dict[str, HealthStatus]


class TooManyHosts(SpicerackError):
    """Raised when there's more hosts than supported passed."""


class UnableToParseOutput(SpicerackError):
    """Raised when there's an error trying to parse etcdctl output."""


class EtcdctlController(RemoteHostsAdapter):
    """Node that is able to run etcdctl and control an etcd cluster."""

    def __init__(self, *, remote_host: RemoteHosts):
        """Init."""
        if len(remote_host.hosts) > 1:
            raise TooManyHosts("EtcdctlController currently only supports running in one node.")

        super().__init__(remote_hosts=remote_host)

        endpoints = f"https://{self._remote_hosts.hosts}:2379"
        cert_file = f"/etc/etcd/ssl/{self._remote_hosts.hosts}.pem"
        ca_file = "/etc/etcd/ssl/ca.pem"
        key_file = f"/etc/etcd/ssl/{self._remote_hosts.hosts}.priv"
        self._base_args = [
            "etcdctl",
            "--endpoints",
            endpoints,
            "--ca-file",
            ca_file,
            "--cert-file",
            cert_file,
            "--key-file",
            key_file,
        ]

    def get_cluster_health(self) -> EtcdClusterHealthStatus:
        """Gets the current etcd cluster health status."""
        args = self._base_args + ["cluster-health"]
        raw_results = self._remote_hosts.run_sync(" ".join(args))
        try:
            result = next(raw_results)[1].message().decode("utf8")
        except StopIteration as e:
            raise UnableToParseOutput("Got no results when trying to retrieve the etcdctl cluster health.") from e

        global_status = None
        members_status = {}
        for line in result.split("\n"):
            line = line.strip()
            if not line:
                continue

            # member <memberid> is <healthy|unhealthy>: got <healthy|unhealthy> result from <member_url>
            # ...
            # cluster is <healthy|unhealthy>
            if line.startswith("cluster is"):
                global_status = HealthStatus[line.rsplit(" ", 1)[-1]]
            else:
                _, member_id, _, raw_health_status, _ = line.split(" ", 4)
                # raw_health_status includes the ':'
                members_status[member_id] = HealthStatus[raw_health_status[:-1]]

        if global_status is None:
            raise UnableToParseOutput(f"Can't find the global cluster status in the cluster-health output: {result}")

        return EtcdClusterHealthStatus(global_status=cast(HealthStatus, global_status), members_status=members_status)

    def get_cluster_info(self) -> Dict[str, Dict[str, SimpleType]]:
        """Gets the current etcd cluster information."""
        args = self._base_args + ["member", "list"]
        raw_results = self._remote_hosts.run_sync(" ".join(args))
        try:
            result = next(raw_results)[1].message().decode("utf8")
        except StopIteration as e:
            raise UnableToParseOutput("Got no results when trying to retrieve the etcdctl members list.") from e

        structured_result = {}
        for line in result.split("\n"):
            if not line.strip():
                continue

            # <memberid>[<status>]: <key>=<value> <key>=<value>...
            # where value might be the string "true" or a stringified int "42"
            # and the '[<status>]' bit might not be there
            # peerURLs and memberid are the only key that seems to be there always
            split_info = [self._to_simple_tuple(elem) for elem in line.split(":", 1)[-1].split()]
            struct_elem: Dict[str, SimpleType] = dict(split_info)

            first_part = line.split(":", 1)[0].strip()
            if "[" in first_part:
                member_id = first_part.split("[", 1)[0]
                status = first_part.split("[", 1)[1][:-1]
            else:
                member_id = first_part
                status = "up"

            struct_elem["member_id"] = member_id
            struct_elem["status"] = status

            if "peerURLs" not in struct_elem:
                raise UnableToParseOutput(
                    "Unable to parse etcdctl output (missing peerURLs for "
                    f"member line):\nParsed: {struct_elem}\nLine: {line}\n"
                    f"Full output: {result}"
                )
            structured_result[member_id] = struct_elem

        return structured_result

    def ensure_node_exists(
        self,
        new_member_fqdn: str,
        member_peer_url: Optional[str] = None,
    ) -> str:
        """Ensure the existance of an etcd member adding it if not present.

        Makes sure that the given new_member_fqdn member exists and is part of
        the etcd cluster, and returns its member id.
        """
        if not member_peer_url:
            member_peer_url = f"https://{new_member_fqdn}:2380"

        before_members = self.get_cluster_info()
        current_entry = self._get_member_or_none(
            members=before_members,
            member_name=new_member_fqdn,
            member_peer_url=member_peer_url,
        )
        extra_args = None

        if current_entry and current_entry["peerURLs"] == member_peer_url:
            logger.info(
                "Skipping addition of member %s as it already exists.",
                new_member_fqdn,
            )
            return cast(str, current_entry["member_id"])

        if current_entry and current_entry["peerURLs"] != member_peer_url:
            logger.info(
                "Updating url for already existing member %s.",
                new_member_fqdn,
            )
            extra_args = [
                "member",
                "update",
                cast(str, current_entry["member_id"]),
                member_peer_url,
            ]

        else:
            extra_args = ["member", "add", new_member_fqdn, member_peer_url]

        self._remote_hosts.run_sync(" ".join(self._base_args + extra_args))

        if not current_entry:
            # unfortunately, etcdctl add does not give the member_id, but only the new
            # name, so we have to diff before and after to find out which one is the
            # new member id
            after_members = self.get_cluster_info()
            new_member_id = list(set(after_members.keys()) - set(before_members.keys()))[0]

        else:
            new_member_id = cast(str, current_entry["member_id"])

        return new_member_id

    def ensure_node_does_not_exist(
        self,
        member_fqdn: str,
    ) -> Optional[str]:
        """Ensure the non existance of an etcd member, removing it if present.

        Makes sure that the given member_fqdn member is not part of the etcd
        cluster, returns its old member id or None if it was not there.
        """
        before_members = self.get_cluster_info()
        current_entry = self._get_member_or_none(
            members=before_members,
            member_name=member_fqdn,
        )

        if not current_entry:
            logger.info(
                "Skipping removal of member %s as it does not exist.",
                member_fqdn,
            )
            return None

        logger.info("Removing etcd member %s.", member_fqdn)
        extra_args = ["member", "remove", str(current_entry["member_id"])]
        self._remote_hosts.run_sync(" ".join(self._base_args + extra_args))
        return str(current_entry["member_id"])

    @staticmethod
    def _to_simple_type(maybe_not_string: str) -> SimpleType:
        """Simple type interpolation, etcdctl member list does not return json."""
        if maybe_not_string == "true":
            return True

        if maybe_not_string == "false":
            return False

        try:
            return int(maybe_not_string)

        except ValueError:
            pass

        return maybe_not_string

    @classmethod
    def _to_simple_tuple(cls, elem: str) -> Tuple[str, SimpleType]:
        elems = elem.split("=", 1)
        if len(elems) != 2:
            raise UnableToParseOutput(f"Malformed element '{elem}', has no '='.")

        return (elems[0], cls._to_simple_type(elems[1]))

    @staticmethod
    def _get_member_or_none(
        members: Dict[str, Dict[str, SimpleType]], member_name: str, member_peer_url: Optional[str] = None
    ) -> Dict[str, SimpleType]:
        return next(
            (
                member
                for member in members.values()
                if (
                    "name" in member
                    and member["name"] == member_name
                    # in case the member is not started, it does not show the name,
                    # just the peer url
                    or "name" not in member
                    and member["peerURLs"] == member_peer_url
                )
            ),
            {},
        )



"""Alerting module."""
import logging
from contextlib import contextmanager
from datetime import timedelta
from typing import Iterator

from spicerack.administrative import Reason
from spicerack.alertmanager import AlertmanagerHosts
from spicerack.icinga import IcingaHosts

logger = logging.getLogger(__name__)


class AlertingHosts:
    """Operate on Alertmanager and Icinga via their APIs."""

    def __init__(
        self,
        alertmanager_hosts: AlertmanagerHosts,
        icinga_hosts: IcingaHosts,
    ) -> None:
        """Initialize the instance.

        Arguments:
            alertmanager_hosts (spicerack.alertmanager.AlertmanagerHosts): the Alertmanager hosts to talk to.
            icinga_hosts (spicerack.icinga.IcingaHosts): the Icinga hosts to talk to.

        """
        self._am = alertmanager_hosts
        self._icinga = icinga_hosts

    @contextmanager
    def downtimed(
        self, reason: Reason, *, duration: timedelta = timedelta(hours=4), remove_on_error: bool = False
    ) -> Iterator[None]:
        """Context manager to perform actions while the hosts are downtimed on Alertmanager and Icinga.

        Arguments:
            reason (spicerack.administrative.Reason): the reason to set for the downtime.
            duration (datetime.timedelta, optional): the length of the downtime period.
            remove_on_error: should the downtime be removed even if an exception was raised.

        Yields:
            None: it just yields control to the caller once Alertmanager and Icinga have
            received the downtime and deletes the downtime once getting back the
            control.

        """
        downtime_id = self.downtime(reason, duration=duration)
        try:
            yield
        except BaseException:
            if remove_on_error:
                self.remove_downtime(downtime_id)
            raise
        else:
            self.remove_downtime(downtime_id)

    def downtime(self, reason: Reason, *, duration: timedelta = timedelta(hours=4)) -> str:
        """Issue a new downtime.

        Arguments:
            reason (spicerack.administrative.Reason): the downtime reason.
            duration (datetime.timedelta): how long to downtime for.

        Returns:
            str: the downtime ID.

        Raises:
            AlertmanagerError: if none of the `alertmanager_urls` API returned a success.
            IcingaError: if there is a problem downtiming the hosts in Icinga.

        """
        self._icinga.downtime(reason, duration=duration)
        return self._am.downtime(reason, duration=duration)

    def remove_downtime(self, downtime_id: str) -> None:
        """Remove a downtime.

        Arguments:
            downtime_id (str): the downtime ID to remove.

        """
        self._icinga.remove_downtime()
        return self._am.remove_downtime(downtime_id)



"""IPMI module tests."""

from subprocess import PIPE, CalledProcessError, CompletedProcess
from unittest import mock

import pytest

from spicerack import ipmi

ENV = {"IPMITOOL_PASSWORD": "password"}
IPMITOOL_BASE = [
    "ipmitool",
    "-I",
    "lanplus",
    "-H",
    "test-mgmt.example.com",
    "-U",
    "root",
    "-E",
]
BOOTPARAMS_OUTPUT = """
Boot parameter version: 1
Boot parameter 5 is valid/unlocked
Boot parameter data: {bootparams}
 Boot Flags :
   - Boot Flag Invalid
   - Options apply to only next boot
   - BIOS PC Compatible (legacy) boot
   - Boot Device Selector : {override}
   - Console Redirection control : System Default
   - BIOS verbosity : Console redirection occurs per BIOS configuration setting (default)
   - BIOS Mux Control Override : BIOS uses recommended setting of the mux at the end of POST
Invalid line
"""
USERLIST_OUTPUT = """ID  Name             Callin  Link Auth  IPMI Msg   Channel Priv Limit
1                    true    false      false      NO ACCESS
2   root             true    true       true       ADMINISTRATOR
3                    true    false      false      NO ACCESS
4                    true    false      false      NO ACCESS
5                    true    false      false      NO ACCESS
6                    true    false      false      NO ACCESS
7                    true    false      false      NO ACCESS
8                    true    false      false      NO ACCESS
9   user_number_9    true    false      false      NO ACCESS
10                   true    false      false      NO ACCESS
11                   true    false      false      NO ACCESS
12                   true    false      false      NO ACCESS
13                   true    false      false      NO ACCESS
14                   true    false      false      NO ACCESS
15                   true    false      false      NO ACCESS
16                   true    false      false      NO ACCESS
"""


class TestIpmi:
    """Test class for the Ipmi class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.fqdn = "test-mgmt.example.com"
        self.ipmi = ipmi.Ipmi(self.fqdn, "password", dry_run=False)
        self.ipmi_dry_run = ipmi.Ipmi(self.fqdn, "password")

    def test_init(self):
        """It should initialize the instance and set the IPMITOOL_PASSWORD environment variable."""
        assert isinstance(self.ipmi, ipmi.Ipmi)

    @mock.patch("spicerack.ipmi.run", return_value=CompletedProcess((), 0, stdout=b"test"))
    def test_command_ok(self, mocked_run):
        """It should execute the IPMI command as expected."""
        assert self.ipmi.command(["test_command"]) == "test"
        mocked_run.assert_called_once_with(IPMITOOL_BASE + ["test_command"], env=ENV, stdout=PIPE, check=True)

    @mock.patch("spicerack.ipmi.run")
    def test_command_dry_run_ok(self, mocked_run):
        """It should not execute the IPMI command if in DRY RUN mode."""
        assert self.ipmi_dry_run.command(["test_command"]) == ""
        assert not mocked_run.called

    @mock.patch("spicerack.ipmi.run")
    def test_command_raise(self, mocked_run):
        """It should raise IpmiError if failed to execute the command."""
        mocked_run.side_effect = CalledProcessError(1, "executed_command")
        with pytest.raises(ipmi.IpmiError, match="Remote IPMI for test-mgmt.example.com failed"):
            self.ipmi.command(["test_command"])

    @mock.patch(
        "spicerack.ipmi.run",
        return_value=CompletedProcess((), 0, stdout=b"Chassis Power is on"),
    )
    def test_check_connection_ok(self, mocked_run):
        """It should check that the connection to the remote IPMI works running a simple command."""
        self.ipmi_dry_run.check_connection()
        mocked_run.assert_called_once_with(
            IPMITOOL_BASE + ["chassis", "power", "status"],
            env=ENV,
            stdout=PIPE,
            check=True,
        )

    @mock.patch(
        "spicerack.ipmi.run",
        return_value=CompletedProcess((), 0, stdout=b"Chassis Power is on\n"),
    )
    def test_power_status_ok(self, mocked_run):
        """It should return the current power status of the target host."""
        status = self.ipmi_dry_run.power_status()
        assert status == "on"
        mocked_run.assert_called_once_with(
            IPMITOOL_BASE + ["chassis", "power", "status"],
            env=ENV,
            stdout=PIPE,
            check=True,
        )

    @mock.patch("spicerack.ipmi.run", return_value=CompletedProcess((), 0, stdout=b"failed"))
    def test_power_status_raise(self, mocked_run):
        """It should raise IpmiError if unable to get the power status."""
        with pytest.raises(ipmi.IpmiError, match="Unexpected chassis status: failed"):
            self.ipmi.power_status()

        assert mocked_run.called

    @pytest.mark.parametrize("status, operation", (("on", "cycle"), ("off", "on")))
    @mock.patch("spicerack.ipmi.run")
    def test_reboot(self, mocked_run, status, operation):
        """It issue the proper reboot command based on the current power status."""
        mocked_run.return_value = CompletedProcess(
            (),
            0,
            stdout=f"Chassis Power is {status}\n".encode(),
        )
        self.ipmi.reboot()
        mocked_run.has_call(
            IPMITOOL_BASE + ["chassis", "power", operation],
            env=ENV,
            stdout=PIPE,
            check=True,
        )

    @mock.patch("spicerack.ipmi.run")
    def test_check_bootparams_ok(self, mocked_run):
        """It should check that the BIOS boot parameters are normal."""
        mocked_run.return_value = CompletedProcess(
            (),
            0,
            stdout=BOOTPARAMS_OUTPUT.format(bootparams="0000000000", override="No override").encode(),
        )
        self.ipmi.check_bootparams()
        mocked_run.assert_called_once_with(
            IPMITOOL_BASE + ["chassis", "bootparam", "get", "5"],
            env=ENV,
            stdout=PIPE,
            check=True,
        )

    @mock.patch("spicerack.ipmi.run")
    def test_check_bootparams_wrong_value(self, mocked_run):
        """It should raise IpmiCheckError if the BIOS boot parameters are not the normal ones."""
        mocked_run.return_value = CompletedProcess(
            (),
            0,
            stdout=BOOTPARAMS_OUTPUT.format(bootparams="0004000000", override="Force PXE").encode(),
        )
        with pytest.raises(
            ipmi.IpmiCheckError,
            match=r"Expected BIOS boot params in \('0000000000', '8000000000', '8000020000'\) got: 0004000000",
        ):
            self.ipmi_dry_run.check_bootparams()

        assert mocked_run.called

    @mock.patch(
        "spicerack.ipmi.run",
        return_value=CompletedProcess((), 0, stdout=b"Boot parameter data"),
    )
    def test_check_bootparams_unable_to_extract(self, mocked_run):
        """It should raise IpmiError if unable to extract the value of the BIOS boot parameters."""
        with pytest.raises(
            ipmi.IpmiError,
            match="Unable to extract value for parameter 'Boot parameter data'",
        ):
            self.ipmi.check_bootparams()

        assert mocked_run.called

    @mock.patch("spicerack.ipmi.run", return_value=CompletedProcess((), 0, stdout=b"Invalid"))
    def test_check_bootparams_missing_label(self, mocked_run):
        """It should raise IpmiError if unable to find the label looked for."""
        with pytest.raises(
            ipmi.IpmiError,
            match="Unable to find the boot parameter 'Boot parameter data'",
        ):
            self.ipmi.check_bootparams()

        assert mocked_run.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    @mock.patch("spicerack.ipmi.run")
    def test_force_pxe_ok(self, mocked_run, mocked_sleep):
        """Should set the PXE boot mode for the next boot."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=b""),
            CompletedProcess(
                (),
                0,
                stdout=BOOTPARAMS_OUTPUT.format(bootparams="0004000000", override="Force PXE").encode(),
            ),
        ]
        self.ipmi.force_pxe()

        assert not mocked_sleep.called
        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["chassis", "bootparam", "set", "bootflag", "force_pxe", "options=reset"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["chassis", "bootparam", "get", "5"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
            ]
        )

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    @mock.patch("spicerack.ipmi.run")
    def test_force_pxe_retried(self, mocked_run, mocked_sleep):
        """Should retry to set the PXE mode on failure."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=b"PXE not set"),
            CompletedProcess(
                (),
                0,
                stdout=BOOTPARAMS_OUTPUT.format(bootparams="0000000000", override="No override").encode(),
            ),
            CompletedProcess((), 0, stdout=b"PXE set"),
            CompletedProcess(
                (),
                0,
                stdout=BOOTPARAMS_OUTPUT.format(bootparams="0004000000", override="Force PXE").encode(),
            ),
        ]
        self.ipmi.force_pxe()
        assert mocked_sleep.called

    @mock.patch("wmflib.decorators.time.sleep", return_value=None)
    @mock.patch("spicerack.ipmi.run")
    def test_force_pxe_dry_run(self, mocked_run, mocked_sleep):
        """Should not raise an exception on dry-run mode when unable to verify the boot parameters."""
        mocked_run.side_effect = [
            CompletedProcess(
                (),
                0,
                stdout=BOOTPARAMS_OUTPUT.format(bootparams="0000000000", override="No override").encode(),
            ),
        ]
        self.ipmi_dry_run.force_pxe()  # should not raise
        assert not mocked_sleep.called

    @mock.patch("spicerack.ipmi.run")
    def test_remove_boot_override_ok(self, mocked_run):
        """Should unset any boot override for the next boot."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=b""),
            CompletedProcess(
                (),
                0,
                stdout=BOOTPARAMS_OUTPUT.format(bootparams="8000000000", override="No override").encode(),
            ),
        ]
        self.ipmi.remove_boot_override()

        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["chassis", "bootparam", "set", "bootflag", "none", "options=reset"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["chassis", "bootparam", "get", "5"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
            ]
        )

    @mock.patch("spicerack.ipmi.run")
    def test_remove_boot_override_fail(self, mocked_run):
        """Should raise an IpmiCheckError exception on failure."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=b"PXE not set"),
            CompletedProcess(
                (),
                0,
                stdout=BOOTPARAMS_OUTPUT.format(bootparams="000400000", override="Force PXE").encode(),
            ),
        ]
        with pytest.raises(ipmi.IpmiCheckError, match="Unable to verify that the boot override was removed."):
            self.ipmi.remove_boot_override()

    @mock.patch("spicerack.ipmi.run")
    def test_remove_boot_override_dry_run(self, mocked_run):
        """Should not raise an exception on dry-run mode when unable to verify the boot parameters."""
        mocked_run.side_effect = [
            CompletedProcess(
                (),
                0,
                stdout=BOOTPARAMS_OUTPUT.format(bootparams="0004000000", override="Force PXE").encode(),
            ),
        ]
        self.ipmi_dry_run.remove_boot_override()  # should not raise

    @mock.patch("spicerack.ipmi.run")
    def test_reset_password_16(self, mocked_run):
        """It should reset the users password and store the password as 16 bytes."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=USERLIST_OUTPUT.encode()),
            CompletedProcess((), 0, stdout=b"Set User Password command successful (user 2)\n"),
            CompletedProcess((), 0, stdout=b"Chassis Power is on"),
        ]
        self.ipmi.reset_password("root", "a" * 16)
        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["user", "list", "1"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["user", "set", "password", "2", "a" * 16, "16"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["chassis", "power", "status"],
                    env={"IPMITOOL_PASSWORD": "a" * 16},
                    stdout=PIPE,
                    check=True,
                ),
            ]
        )
        assert self.ipmi.env["IPMITOOL_PASSWORD"] == "a" * 16

    @mock.patch("spicerack.ipmi.run")
    def test_reset_password_20(self, mocked_run):
        """It should reset the users password and store the password as 20 bytes."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=USERLIST_OUTPUT.encode()),
            CompletedProcess((), 0, stdout=b"Set User Password command successful (user 2)\n"),
            CompletedProcess((), 0, stdout=b"Chassis Power is on"),
        ]
        self.ipmi.reset_password("root", "a" * 17)
        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["user", "list", "1"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["user", "set", "password", "2", "a" * 17, "20"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["chassis", "power", "status"],
                    env={"IPMITOOL_PASSWORD": "a" * 17},
                    stdout=PIPE,
                    check=True,
                ),
            ]
        )
        assert self.ipmi.env["IPMITOOL_PASSWORD"] == "a" * 17

    @mock.patch("spicerack.ipmi.run")
    def test_reset_password_not_root(self, mocked_run):
        """It should reset the users password."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=USERLIST_OUTPUT.encode()),
            CompletedProcess((), 0, stdout=b"Set User Password command successful (user 9)\n"),
        ]
        self.ipmi.reset_password("user_number_9", "a" * 16)
        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["user", "list", "1"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["user", "set", "password", "9", "a" * 16, "16"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
            ]
        )

    @mock.patch("spicerack.ipmi.run")
    def test_reset_password_dryrun(self, mocked_run):
        """It should not reset the users password."""
        mocked_run.return_value = CompletedProcess((), 0, stdout=USERLIST_OUTPUT.encode())
        self.ipmi_dry_run.reset_password("root", "a" * 16)
        mocked_run.called_once_with(IPMITOOL_BASE + ["user", "list", "1"], env=ENV, stdout=PIPE, check=True)

    @mock.patch("spicerack.ipmi.run")
    def test_reset_password_fail_command(self, mocked_run):
        """It should fail the password reset command."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=USERLIST_OUTPUT.encode()),
            CompletedProcess((), 0, stdout=b"Fail password reset\n"),
        ]
        with pytest.raises(ipmi.IpmiError, match="Password reset failed for username: user_number_9"):
            self.ipmi.reset_password("user_number_9", "a" * 16)
        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["user", "list", "1"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["user", "set", "password", "9", "a" * 16, "16"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
            ]
        )
        assert self.ipmi.env["IPMITOOL_PASSWORD"] == "password"

    @mock.patch("spicerack.ipmi.run")
    def test_reset_password_connection_test(self, mocked_run):
        """It should fail the check connection command."""
        mocked_run.side_effect = [
            CompletedProcess((), 0, stdout=USERLIST_OUTPUT.encode()),
            CompletedProcess((), 0, stdout=b"Set User Password command successful (user 2)\n"),
            CompletedProcess((), 0, stdout=b"Failed connection test"),
        ]
        with pytest.raises(ipmi.IpmiError, match="Password reset failed for username: root"):
            self.ipmi.reset_password("root", "a" * 16)
        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["user", "list", "1"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["user", "set", "password", "2", "a" * 16, "16"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["chassis", "power", "status"],
                    env={"IPMITOOL_PASSWORD": "a" * 16},
                    stdout=PIPE,
                    check=True,
                ),
            ]
        )
        assert self.ipmi.env["IPMITOOL_PASSWORD"] == "password"

    @mock.patch("spicerack.ipmi.run")
    def test_reset_nonexistent_username(self, mocked_run):
        """It should raise IpmiError as the username will not be found."""
        with pytest.raises(ipmi.IpmiError, match="Unable to find ID for username: nonexistent"):
            self.ipmi.reset_password("nonexistent", "a" * 16)
        mocked_run.assert_has_calls(
            [
                mock.call(
                    IPMITOOL_BASE + ["user", "list", "1"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
                mock.call(
                    IPMITOOL_BASE + ["user", "list", "2"],
                    env=ENV,
                    stdout=PIPE,
                    check=True,
                ),
            ],
            any_order=True,
        )

    def test_reset_password_bad_username(self):
        """It should raise IpmiError is username is empty."""
        with pytest.raises(ipmi.IpmiError, match="Username can not be an empty string"):
            self.ipmi.reset_password("", "a" * 16)

    def test_reset_password_short_password(self):
        """It should raise IpmiError as password is less then 16 bytes."""
        with pytest.raises(ipmi.IpmiError, match="New passwords must be 16 bytes minimum"):
            self.ipmi.reset_password("root", "a" * 15)

    def test_reset_password_long_password(self):
        """It should raise IpmiError as password is larger then 20 bytes."""
        with pytest.raises(ipmi.IpmiError, match="New passwords is greater then the 20 byte limit"):
            self.ipmi.reset_password("root", "a" * 21)
